/*
 * University of Illinois Open Source License
 * Copyright 2014-2015 Luthey-Schulten Group,
 * All rights reserved.
 * 
 * Developed by: Luthey-Schulten Group
 * 			     University of Illinois at Urbana-Champaign
 * 			     http://www.scs.uiuc.edu/~schulten
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the Software), to deal with 
 * the Software without restriction, including without limitation the rights to 
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies 
 * of the Software, and to permit persons to whom the Software is furnished to 
 * do so, subject to the following conditions:
 * 
 * - Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimers.
 * 
 * - Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimers in the documentation 
 * and/or other materials provided with the distribution.
 * 
 * - Neither the names of the Luthey-Schulten Group, University of Illinois at
 * Urbana-Champaign, nor the names of its contributors may be used to endorse or
 * promote products derived from this Software without specific prior written
 * permission.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL 
 * THE CONTRIBUTORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR 
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR 
 * OTHER DEALINGS WITH THE SOFTWARE.
 *
 * Author(s): John Cole
 */




#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <cstdlib>
#include <cstdio>
#include <cuda_runtime_api.h>
#include <cuda_runtime.h>
#include <cuda.h>


using namespace std;


/*          |  
 *   #####  | z
 *  #####   |
 *  #####   /
 *  #####  / y
 *  _____ /
 *    x
 * 
 * my block will be each x slice.  each block will have y * z threads.
 * i will load to shared the elements x(i-1)--x(i+1) * y * z
 * so that i dont have to look for data as im going.
 * i will loop from x(1) to x(Xmax - 1) leaving the ends undone.
 * i wont do the ends of y or z either.
 * 
 */
#define idx(i, j, k, yl, zl) ((i + 1) * (zl + 2) * (yl + 2) + (j + 1) * (zl + 2) + (k + 1))
#define getX(index, yl, zl) (index / ((yl + 2) * (zl + 2)))
#define getY(index, yl, zl) ((index % ((yl + 2) * (zl + 2))) / (zl + 2))
#define getZ(index, yl, zl) (((index % ((yl + 2) * (zl + 2))) % (zl + 2)) - 1)

#define defaultThreads 512
#define fbaThreads 256
#define diffThreads 256
#define syncTCLwithCLThreads 256
//#define chemReactThreads 192 for compute 2.0, 256 for compute 3.0
#define expandCellsThreads 256

#define maxVolFracTolerance 0.01f

//========================================universal host functions========================================
/// @brief Check if a file was opened successfully.
/// @param file A pointer to a file
/// @param str A the name of the file
/// @param code Exit code if failed to open (-1 by default)
/// @return Void
void checkFileOpenedSuccessfully(ios *file, string str, int code = -1) {
	if(!file->good()) {
			cerr << "ERROR: could not open " << str << "; exiting..." << endl;
			exit(code);
	}
}

/// @brief Get the rest of a line in a file after an identifier.
/// @param filename The name of the file
/// @param identifier A string that identifies the line to be returned
/// @return A string containing the rest of the line from the file after the identifier
string getStringAfterIdentifier(string filename, string identifier){
	ifstream inputFile;
	inputFile.open(filename.c_str());
	checkFileOpenedSuccessfully(&inputFile, filename);
	
	string thisLine;
	int start;
	bool eof;
	while(true){
		eof = (bool)((getline(inputFile, thisLine)).eof());
		if(eof){
			inputFile.close();
			return "";
		}
		start = thisLine.find(identifier);
		if(start != string::npos){
			inputFile.close();
			return thisLine.substr((start + identifier.length()), string::npos);
		}
	}
}

/// @brief Split a string into a vector of strings at a delimiter.
/// @param input The string to be split
/// @param delimiter The delimiter
/// @return A vector of substrings
vector< string > split(string input, string delimiter){
	//split up this string:
	
	/*
	sstream strsplitter(input);
	while(strsplitter.good())
	{
		string str;
		strsplitter >> str;
		temp.push_back(str);
	}
	*/
	
	
	vector< int > temp;
	for(int i = 0; i < input.length(); i++){
		if(input.substr(i, 1).compare(delimiter) == 0){
			temp.push_back(i);
		}
	}
	vector< string > substrings;
	substrings.push_back(input.substr(0, temp[0]));
	for(int i = 0; i < temp.size() - 1; i++){
		substrings.push_back(input.substr((temp[i] + 1), (temp[i + 1] - temp[i] - 1)));
	}
	substrings.push_back(input.substr(temp.back() + 1));
	
	return(substrings);
}

/// @brief Get the maxT parameter from an input file.
/// @param filename The name of the input file
/// @return maxT as a float
float getMaxT(string filename){
	string maxt = getStringAfterIdentifier(filename, "maxT:\t");
	if (maxt.length() > 0){
		return((float)atof(maxt.c_str()));
	}
	//return -1.0;
	cerr << "ERROR: did not find maxT in input file..." << endl;
	exit(-1);
}

/// @brief Get the pullbackT parameter (the time interval at which a simulation state is written to file) from an input file.
/// @param filename The name of the input file
/// @return pullbackT as a float
float getPullbackT(string filename){
	string pullbackT = getStringAfterIdentifier(filename, "pullbackT:\t");
	if (pullbackT.length() > 0){
		return((float)atof(pullbackT.c_str()));
	}
	//return -1.0;
	cerr << "ERROR: did not find pullbackT in input file..." << endl;
	exit(-1);
}

/// @brief Get the growT parameter (the time interval at which the cells are allowed to grow and expand) from an input file.
/// @param filename The name of the input file
/// @return growT as a float
float getGrowT(string filename){
	string growT = getStringAfterIdentifier(filename, "growT:\t");
	if (growT.length() > 0){
		return((float)atof(growT.c_str()));
	}
	//return -1.0;
	cerr << "ERROR: did not find growT in input file..." << endl;
	exit(-1);
}

/// @brief Get the checkpointT parameter (the time interval at which the simulation state is archived as a checkpoint) from an input file.
/// @param filename The name of the input file
/// @return checkpointT as a float
float getCheckpointT(string filename){
	string checkpointT = getStringAfterIdentifier(filename, "checkpointT:\t");
	if (checkpointT.length() > 0){
		return((float)atof(checkpointT.c_str()));
	}
	return -1.0;
}

/// @brief Get the chemEqT parameter (the time interval at which the diffusing species have equilibrated between growth events) from an input file.
/// @param filename The name of the input file
/// @return chemEqT as a float
float getChemEqT(string filename){
	string chemEqT = getStringAfterIdentifier(filename, "chemEqT:\t");
	if (chemEqT.length() > 0){
		return((float)atof(chemEqT.c_str()));
	}
	//return -1.0;
	cerr << "ERROR: did not find chemEqT in input file..." << endl;
	exit(-1);
}

/// @brief Get the initial chemEqT parameter (the time interval over which the diffusing species are allowed to equilibrat prior to simulation launch) from an input file.
/// @param filename The name of the input file
/// @return initialChemEqT as a float
float getInitialChemEqT(string filename){
	string initialChemEqT = getStringAfterIdentifier(filename, "initialChemEqT:\t");
	if (initialChemEqT.length() > 0){
		return((float)atof(initialChemEqT.c_str()));
	}
	return 0.0; //if not found, return 0.0.
}

/// @brief Get the cellMass parameter (the average mass of a cell in the simulation) from an input file.
/// @param filename The name of the input file
/// @return cellMass as a float
float getCellMass(string filename){
	string cellMass = getStringAfterIdentifier(filename, "cellMass:\t");
	if (cellMass.length() > 0){
		return((float)atof(cellMass.c_str()));
	}
	//return -1.0;
	cerr << "ERROR: did not find cellMass in input file..." << endl;
	exit(-1);
}

/// @brief Get the cellVolume parameter (the average volume of a cell in the simulation) from an input file.
/// @param filename The name of the input file
/// @return cellVolume as a float
float getCellVolume(string filename){
	string cellVolume = getStringAfterIdentifier(filename, "cellVolume:\t");
	if (cellVolume.length() > 0){
		return((float)atof(cellVolume.c_str()));
	}
	//return -1.0;
	cerr << "ERROR: did not find cellVolume in input file..." << endl;
	exit(-1);
}

/// @brief Get the lattice size (number of sites) in the X-direction from an input file.
/// @param filename The name of the input file
/// @return xl as an int
int getLx(string filename){
	string s = getStringAfterIdentifier(filename, "xl:\t");
	if (s.length() > 0){
		return((int)atoi(s.c_str()));
	}
	//return -1;
	cerr << "ERROR: did not find xl in input file..." << endl;
	exit(-1);
}

/// @brief Get the lattice size (number of sites) in the Y-direction from an input file.
/// @param filename The name of the input file
/// @return yl as an int
int getLy(string filename){
	string s = getStringAfterIdentifier(filename, "yl:\t");
	if (s.length() > 0){
		return((int)atoi(s.c_str()));
	}
	//return -1;
	cerr << "ERROR: did not find yl in input file..." << endl;
	exit(-1);
}

/// @brief Get the lattice size (number of sites) in the Z-direction from an input file.
/// @param filename The name of the input file
/// @return zl as an int
int getLz(string filename){
	string s = getStringAfterIdentifier(filename, "zl:\t");
	if (s.length() > 0){
		return((int)atoi(s.c_str()));
	}
	//return -1;
	cerr << "ERROR: did not find zl in input file..." << endl;
	exit(-1);
}

/// @brief Get the lattice spacing (in meters) from an input file.
/// @param filename The name of the input file
/// @return lambda as a float
float getLambda(string filename){
	string s = getStringAfterIdentifier(filename, "lambda:\t");
	if (s.length() > 0){
		return((float)atof(s.c_str()));
	}
	//return -1.0;
	cerr << "ERROR: did not find lambda in input file..." << endl;
	exit(-1);
}

/// @brief Get the FBA timestep from an input file.
/// @param filename The name of the input file
/// @return fbaTau as a float
float getFbaTau(string filename){
	string s = getStringAfterIdentifier(filename, "fbaTau:\t");
	if (s.length() > 0){
		return((float)atof(s.c_str()));
	}
	//return -1.0;
	cerr << "ERROR: did not find fbaTau in input file..." << endl;
	exit(-1);
}

/// @brief Get the agar height (in number of lattice sites) from an input file.
/// @param filename The name of the input file
/// @return agarHeight as an int
int getAgarHeight(string filename){
	string s = getStringAfterIdentifier(filename, "agarHeight:\t");
	if (s.length() > 0){
		return((int)atoi(s.c_str()));
	}
	//return -1;
	cerr << "ERROR: did not find agarHeight in input file..." << endl;
	exit(-1);
}

/// @brief Get the maximum cell volume fraction parameter from an input file.
/// @param filename The name of the input file
/// @return maxVolFrac as a float
float getMaxVolFrac(string filename){
	string s = getStringAfterIdentifier(filename, "maxVolFrac:\t");
	if (s.length() > 0){
		return((float)atof(s.c_str()));
	}
	//return -1.0;
	cerr << "ERROR: did not find maxVolFrac in input file..." << endl;
	exit(-1);
}


//===========================================fba model and related functions======================================
/// @brief Get the number of input metabolite upper bounds for an FBA table from an FBA table file.
/// @param filename The name of the FBA table file
/// @return nInputs as an int
int getNinputs(string filename){
	string s = getStringAfterIdentifier(filename, "nInputs:\t");
	if (s.length() > 0){
		return((int)atoi(s.c_str()));
	}
	//return -1;
	cerr << "ERROR: FBA table file, " << filename << ", is no good; missing nInputs..." << endl;
	exit(-2);
}

/// @brief Get the number of output metabolite fluxes for an FBA table from an FBA table file.
/// @param filename The name of the FBA table file
/// @return nOutputs as an int
int getNoutputs(string filename){
	string s = getStringAfterIdentifier(filename, "nOutputs:\t");
	if (s.length() > 0){
		return((int)atoi(s.c_str()));
	}
	//return -1;
	cerr << "ERROR: FBA table file, " << filename << ", is no good; missing nOutputs..." << endl;
	exit(-2);
}

/// @brief Get the indexes of the metabolites that correspond to the inputs of the FBA table from an FBA table file.
/// @param filename The name of the FBA table file
/// @return inputIndexes as a string
string getInputIndexes(string filename){
	string s = getStringAfterIdentifier(filename, "inputIndexes:\t");
	if (s.length() > 0){
		return s;
	}
	//return -1;
	cerr << "ERROR: FBA table file, " << filename << ", is no good; missing inputIndexes..." << endl;
	exit(-2);	
}

/// @brief Get the indexes of the metabolites that correspond to the outputs of the FBA table from an FBA table file.
/// @param filename The name of the FBA table file
/// @return outputIndexes as a string
string getOutputIndexes(string filename){
	string s = getStringAfterIdentifier(filename, "outputIndexes:\t");
	if (s.length() > 0){
		return s;
	}
	//return -1;
	cerr << "ERROR: FBA table file, " << filename << ", is no good; missing outputIndexes..." << endl;
	exit(-2);	
}

/// @brief Get the numbers of upper bounds tabled for each metabolite of the FBA table from an FBA table file.
/// @param filename The name of the FBA table file
/// @return numberOfTabledValues as a string
string getNumberOfTabledValues(string filename){
	string s = getStringAfterIdentifier(filename, "numberOfTabledValues:\t");
	if (s.length() > 0){
		return s;
	}
	//return -1;
	cerr << "ERROR: FBA table file, " << filename << ", is no good; missing numberOfTabledValues..." << endl;
	exit(-2);	
}

/// @brief Get the difference in consecutive upper bounds for each of the input metabolites of the FBA table from an FBA table file.
/// @param filename The name of the FBA table file
/// @return inputStride as a string
string getInputStride(string filename){
	string s = getStringAfterIdentifier(filename, "inputStride:\t");
	if (s.length() > 0){
		return s;
	}
	//return -1;
	cerr << "ERROR: FBA table file, " << filename << ", is no good; missing inputStride..." << endl;
	exit(-2);	
}

/// @brief Store the FBA table in an FBA table file into an array.
/// @param filename The name of the FBA table file
/// @param fbaTable Pointer to an FBA Table array (floats)
/// @param N The number of elements in the table
/// @return Void
void getFbaTable(string filename, float *fbaTable, int N){
	ifstream inputFile;
	inputFile.open(filename.c_str());
	
	string thisLine;
	string testString = "fbaTable:";
	bool copy = false;
	bool eof;
	int counter = 0;
	int start;
	
	vector< string > thisSplitLine;
	
	while(true){
		eof = (bool)((getline(inputFile, thisLine)).eof());
		if(eof){
			break;
		}
		if(counter == N){
			break;
		}
		
		if(copy){
			thisSplitLine = split(thisLine, "\t");
			for(int i = 0; i < thisSplitLine.size(); ++i){
				fbaTable[counter] = (float)atof((thisSplitLine[i]).c_str());
				++counter;
			}
		}
		
		start = thisLine.find(testString);
		if(start != string::npos){
			copy = true;
		}
	}
	inputFile.close();
}

/// @brief fbaModel struct.
/// This is a struct that contains an FBA table and the parameters needed to index into it.
struct fbaModel{
	int nInputs;
	int nOutputs;
	int inputIndexes[10]; //ten max
	int outputIndexes[11]; //always one larger than the number of inputs, so that the last one is growth rate. 
	int numberOfTabledValues[10];
	float inputStride[10]; //the amount by which each fba input changes
	int indexMultipliers[10];
	float *fbaTable;
	float *cFbaTable;
	
// 	fbaModel(){
// 		
// 	}
	
	/// @brief build the FBA table using an FBA table file.
	/// @param filename The name of the FBA table file
	/// @return Void
	void buildFbaModel(string inputFilename){
		//malloc and cudaMalloc etc.
		nInputs = getNinputs(inputFilename);
		nOutputs = getNoutputs(inputFilename); //n outputs should be some number, most likely nInputs, plus 1 for gr.
		
		string inpIdx = getInputIndexes(inputFilename);
		vector< string > inpIdxV = split(inpIdx, "\t");
		for(int i = 0; i < nInputs; ++i){
			inputIndexes[i] = (int)atoi(inpIdxV[i].c_str());
		}
		
		string outIdx = getOutputIndexes(inputFilename);
		vector< string > outIdxV = split(outIdx, "\t");
		for(int i = 0; i < nOutputs - 1; ++i){
			outputIndexes[i] = (int)atoi(outIdxV[i].c_str());
		}
		
		string numVals = getNumberOfTabledValues(inputFilename);
		vector< string > numValsV = split(numVals, "\t");
		for(int i = 0; i < nInputs; ++i){
			numberOfTabledValues[i] = (int)atoi(numValsV[i].c_str());
		}
		
		string stride = getInputStride(inputFilename);
		vector< string > strideV = split(stride, "\t");
		for(int i = 0; i < nInputs; ++i){
			inputStride[i] = (float)atof(strideV[i].c_str());
		}
		
		//table up the index multipliers
		for(int i = 0; i < nInputs - 1; ++i){
			indexMultipliers[i] = 1;
			for(int j = i + 1; j < nInputs; ++j){
				indexMultipliers[i] *= numberOfTabledValues[j];
			}
		}
		indexMultipliers[nInputs - 1] = 1;
		
		// take care of the table
		int N = nOutputs;
		for(int i = 0; i < nInputs; ++i){
			N *= numberOfTabledValues[i];
		}
		fbaTable = (float *) malloc(sizeof(float) * N);
		getFbaTable(inputFilename, fbaTable, N);
		
		cudaMalloc((void **)&cFbaTable, sizeof(float) * N);
		
		cudaMemcpy(cFbaTable, fbaTable, sizeof(float) * N, cudaMemcpyHostToDevice);
		
	}
};


//=====================================cells structs and related functions:================================
//BETTER TO MAKE REACTIONS PART OF CELLS, LIKE EACH CELL COULD HAVE ITS OWN RECTIONS STRUCT
/// @brief Cell lattice struct.
/// This contains a lattice of the local volume fraction of a single cell type, as well as the FBA table the cell uses.
struct cellLattice{
	
	float *volumeFraction;
	float *cVolumeFraction;
	float *tempVolumeFraction;
	float *cTempVolumeFraction;
	
	float* fbaOuts[11];
	float* cFbaOuts[11];
	
	fbaModel fba;
	int starterFlag; //is this the first cell at the beginning?
	
// 	cellLattice(){
// 		
// 	}
	
	/// @brief Initialize the starting geometry of a cell type using the an arbitrary volume fraction file.
	/// @param volumeFraction Pointer to a cell's volume fraction lattice
	/// @param arbitraryVolumeFractionFilename Name of the arbitrary volume fraction file
	/// @param ly Length of the lattice in the Y-direction
	/// @param lz Length of the lattice in the Z-direction
	/// @return Void
	void readArbitraryVolumeFractionFilename(float *volumeFraction, string arbitraryVolumeFractionFilename, int ly, int lz){
		
		ifstream inputFile;
		inputFile.open(arbitraryVolumeFractionFilename.c_str());
		
		string thisLine;
		string testString = "arbitraryVolumeFractions:";
		bool copy = false;
		bool eof;
		int counter = 0;
		int start;
		
		int x;
		int y;
		int z;
		float rho;
		int arrOfTabs [3];
		int ct;
		
		while(true){
			eof = (bool)((getline(inputFile, thisLine)).eof());
			if(eof){
				break;
			}
// 			if(counter == N){
// 				break;
// 			}
			
			if(copy){
				ct = 0;
				//get array of tabs
				for(int i = 0; i < thisLine.length(); ++i){
					if(thisLine.compare(i, 1, "\t") == 0){
						arrOfTabs[ct] = i;
						++ct;
					}
				}
				x = (int)atoi((thisLine.substr(0, arrOfTabs[0])).c_str());
				y = (int)atoi((thisLine.substr(arrOfTabs[0]+1, arrOfTabs[1]-(arrOfTabs[0]+1))).c_str());
				z = (int)atoi((thisLine.substr(arrOfTabs[1]+1, arrOfTabs[2]-(arrOfTabs[1]+1))).c_str());
				rho = (float)atof((thisLine.substr(arrOfTabs[2]+1, thisLine.length() - arrOfTabs[2])).c_str());
				
				volumeFraction[idx(x, y, z, ly, lz)] = rho;
				
				++counter;
			}
			
			start = thisLine.find(testString);
			if(start != string::npos){
				copy = true;
			}
		}
		inputFile.close();
	}
	
	/// @brief Build a cell lattice struct.
	/// @param f An FBA table struct
	/// @param x Length of the lattice in the X-direction
	/// @param y Length of the lattice in the Y-direction
	/// @param z Length of the lattice in the Z-direction
	/// @param lambda Lattice Spacing (in meters)
	/// @param start Flag (0 or 1) indicating if this cell type is present at the start of a simulation
	/// @param arbitraryVolumeFractionFilename Name of the arbitrary volume fraction file (can be '' if none)
	/// @param startX Starting X location (in units of the lattice spacing) for the cell (if start = 1 and arbitraryVolumeFractionFilename = '')
	/// @param startY Starting Y location (in units of the lattice spacing) for the cell (if start = 1 and arbitraryVolumeFractionFilename = '')
	/// @param startNumber Number of cells initially in the simulation (if start = 1 and arbitraryVolumeFractionFilename = '')
	/// @param agarHeight Height of the agar (in units of the lattice spacing)
	/// @param cellV Average cell volume
	/// @return Void
	void buildCellLattice(fbaModel f, int x, int y, int z, float lambda, int start, string arbitraryVolumeFractionFilename, int startX, int startY, int startNumber, int agarHeight, float cellV){
		fba = f;
		starterFlag = start;
		
		//malloc and cudamalloc etc.
		int N = (x + 2) * (y + 2) * (z + 2);
		
		volumeFraction = (float *) malloc(sizeof(float) * N);
		tempVolumeFraction = (float *) malloc(sizeof(float) * N);
		
		for(int i = 0; i < N; ++i){
			volumeFraction[i] = 0.0f;
			tempVolumeFraction[i] = 0.0f;
		}
		
		if(start > 0){
			if(arbitraryVolumeFractionFilename.length() > 0){
				readArbitraryVolumeFractionFilename(volumeFraction, arbitraryVolumeFractionFilename, y, z);
			}
			else{
				int center = idx(startX, startY, (agarHeight + 1), y, z);
				volumeFraction[center] = cellV * startNumber / (lambda * lambda * lambda);
				tempVolumeFraction[center] = cellV * startNumber / (lambda * lambda * lambda);
			}
		}
		
		cudaMalloc((void **)&cVolumeFraction, sizeof(float) * N);
		cudaMalloc((void **)&cTempVolumeFraction, sizeof(float) * N);
		
		cudaMemcpy(cVolumeFraction, volumeFraction, sizeof(float) * N, cudaMemcpyHostToDevice);
		cudaMemcpy(cTempVolumeFraction, tempVolumeFraction, sizeof(float) * N, cudaMemcpyHostToDevice);
		
		for(int i = 0; i < fba.nOutputs; ++i){
			fbaOuts[i] = (float *) malloc(sizeof(float) * N);
			cudaMalloc((void **)&cFbaOuts[i], sizeof(float) * N);
			for(int j = 0; j < N; ++j){
				fbaOuts[i][j] = 0.0f;
			}
			cudaMemcpy(cFbaOuts[i], fbaOuts[i], sizeof(float) * N, cudaMemcpyHostToDevice);
		}
	}
};

/// @brief Cell lattices struct.
/// This is a container struct for all the cell lattice structs in the simulation.  It also stores to total volume fraction lattice
struct cellLattices{
	int *typeFlags;
	int *cTypeFlags;
	float *totalVolumeFraction;
	float *cTotalVolumeFraction;
	
	int numberCells;
	cellLattice lattices[10]; //at most ten.
	
	float maxVolFraction;
	float conversionFactorConcPerSecondToMmolPerGDwtPerHr;
	int xl;
	int yl;
	int zl;
	float lambda;
	float fbaTau;
	int agarHeight;
	int N;
	int isRelaxed;
	
	cellLattices(){
		
	}
	
	/// @brief Build the cellLattices struct.
	/// @param mvf Maximum volume fraction parameter
	/// @param x Length of the lattice in the X-direction
	/// @param y Length of the lattice in the Y-direction
	/// @param z Length of the lattice in the Z-direction
	/// @param l Lattice Spacing (in meters)
	/// @param t FBA timestep
	/// @param ah Agar height in units of the lattice spacing
	/// @param cellV Average cell volume
	/// @param cellM Average cell mass
	/// @return Void
	void buildCellLattices(float mvf, int x, int y, int z, float l, float t, int ah, float cellV, float cellM){
		//create the typeFlags and volFraction lattices, and their copies on device.
		numberCells = 0;
		maxVolFraction = mvf;
		xl = x;
		yl = y;
		zl = z;
		lambda = l;
		fbaTau = t;
		agarHeight = ah;
		conversionFactorConcPerSecondToMmolPerGDwtPerHr = 1000.0 * (lambda * lambda * lambda) //molar/s -> mol/s
														* 1000.0 //mol/s -> mmol/s
														* (60.0 * 60.0 / t) //mmol/s -> mmol/hr
														/ (lambda * lambda * lambda / cellV) // change to cells per subvolume assuming one cell is 1e-18 m^3
														/ cellM; //change to grams dry weight
		
		N = (xl + 2) * (yl + 2) * (zl + 2);
		isRelaxed = 0;
		
		typeFlags = (int *) malloc(sizeof(int) * N);
		totalVolumeFraction = (float*) malloc(sizeof(float) * N);
		
		for(int i = 0; i < N; ++i){
			if(getZ(i, yl, zl) > agarHeight){
				typeFlags[i] = 0;
			}
			else{
				typeFlags[i] = 1;
			}
			totalVolumeFraction[i] = 0.0f;
		}
		
		cudaMalloc((void **)&cTotalVolumeFraction, sizeof(float) * N);
		cudaMalloc((void **)&cTypeFlags, sizeof(int) * N);
		
		cudaMemcpy(cTotalVolumeFraction, totalVolumeFraction, sizeof(float) * N, cudaMemcpyHostToDevice);
		cudaMemcpy(cTypeFlags, typeFlags, sizeof(int) * N, cudaMemcpyHostToDevice);
	}
	
	void add(cellLattice cL){
		lattices[numberCells] = cL;
		numberCells++;
	}
};

/// @brief Check the input file to see if the next sequential cell type is defined.
/// @param filename Name of the input file
/// @param cells Container struct of all the current cell types defined.
/// @return True if the N+1 cell type is defined in the input file (when there are currently N cell types loaded) and False if not.
bool isCellHere(string filename, cellLattices cells){
	ifstream inputFile;
	inputFile.open(filename.c_str());
	
	char buffer [4096];
	int n = sprintf(buffer, "cellType%d:\t", cells.numberCells);
	string testString = string(buffer, n);
	
	string thisLine;
	int start;
	
	bool eof;
	
	while(true){
		eof = (bool)((getline(inputFile, thisLine)).eof());
		if(eof){
			inputFile.close();
			return false;
		}
		
		start = thisLine.find(testString);
		if(start != string::npos){
			inputFile.close();
			return true; //as well as the FBA table the cell uses
		}
	}
}

/// @brief Get the name of the fbaTable file for the most recently loaded cell type.
/// @param filename Name of the input file
/// @param cells Container struct of all the current cell types defined.
/// @return The name of the fbaTable file for the last-loaded cell type
string getFbaFilename(string filename, cellLattices cells){
	char buffer [4096];
	int n = sprintf(buffer, "cellType%d fbaFilename:\t", cells.numberCells);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if (s.length() > 0){
		return s;
	}
	//return -1;
	cerr << "ERROR: " << testString << " not found..." << endl;
	exit(-2);	//as well as the FBA table the cell uses
}

/// @brief Get the flag for the most recently loaded cell type to determine if it is present at the start of the simulation.
/// @param filename Name of the input file
/// @param cells Container struct of all the current cell types defined.
/// @return 1 if the most recently loaded cell type is present at the start of the simulation, 0 otherwise
int getStarterFlag(string filename, cellLattices cells){
	char buffer [4096];
	int n = sprintf(buffer, "cellType%d starterFlag:\t", cells.numberCells);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return 0;
}

/// @brief Get the name of the arbitraryVolumeFraction file for the most recently loaded cell type.
/// @param filename Name of the input file
/// @param cells Container struct of all the current cell types defined.
/// @return The name of the arbitraryVolumeFraction file for the last-loaded cell type
string getArbitraryVolumeFractionFilename(string filename, cellLattices cells){
	
	char buffer [4096];
	int n = sprintf(buffer, "cellType%d arbitraryVolumeFractionFilename:\t", cells.numberCells);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		cout << testString << s << endl;
	}
	return s;
}

/// @brief Get the starting X-location for the most recently loaded cell type.
/// @param filename Name of the input file
/// @param cells Container struct of all the current cell types defined.
/// @return The starting X-location for the last-loaded cell type
int getStartX(string filename, cellLattices cells){
	char buffer [4096];
	int n = sprintf(buffer, "cellType%d initX:\t", cells.numberCells);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return -1;
}

/// @brief Get the starting Y-location for the most recently loaded cell type.
/// @param filename Name of the input file
/// @param cells Container struct of all the current cell types defined.
/// @return The starting Y-location for the last-loaded cell type
int getStartY(string filename, cellLattices cells){
	char buffer [4096];
	int n = sprintf(buffer, "cellType%d initY:\t", cells.numberCells);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return -1;
}

/// @brief Get the starting number of cells for the most recently loaded cell type.
/// @param filename Name of the input file
/// @param cells Container struct of all the current cell types defined.
/// @return The starting number of cells for the last-loaded cell type
int getStartNumber(string filename, cellLattices cells){
	char buffer [4096];
	int n = sprintf(buffer, "cellType%d initNumber:\t", cells.numberCells);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return -1;
}

/// @brief Swap normal and temporary volume fraction pointers for all cell types on the host.
/// @param cells Container struct of all the current cell types defined.
/// @return Void
void swapCellLattices_host(cellLattices *cells){
	for(int i = 0; i < cells->numberCells; ++i){
		float *temp = cells->lattices[i].cTempVolumeFraction;
		cells->lattices[i].cTempVolumeFraction = cells->lattices[i].cVolumeFraction;
		cells->lattices[i].cVolumeFraction = temp;
	}
}

/// @brief Swap normal and temporary volume fraction pointers for a single cell type on the host.
/// @param cells Container struct of all the current cell types defined.
/// @param cellIdx Index of the cell type whose pointers will be swapped.
/// @return Void
void swapACellLattice_host(cellLattices *cells, int cellIdx){
	float *temp = cells->lattices[cellIdx].cTempVolumeFraction;
	cells->lattices[cellIdx].cTempVolumeFraction = cells->lattices[cellIdx].cVolumeFraction;
	cells->lattices[cellIdx].cVolumeFraction = temp;
}

/// @brief Copy device side cell data to host side cell data arrays.
/// @param cells Container struct of all the current cell types defined.
/// @return Void
void cellMemCopyBack(cellLattices *cells){
	// total volume fraction first:
	cudaMemcpy(cells->totalVolumeFraction, cells->cTotalVolumeFraction, cells->N * sizeof(float), cudaMemcpyDeviceToHost);
	cudaMemcpy(cells->typeFlags, cells->cTypeFlags, cells->N * sizeof(int),  cudaMemcpyDeviceToHost);
	
	//do the individual cell types;
	for(int cell = 0; cell < cells->numberCells; ++cell){
		//take back vol frac and fba indexes;
		cudaMemcpy(cells->lattices[cell].volumeFraction, cells->lattices[cell].cVolumeFraction, cells->N * sizeof(float), cudaMemcpyDeviceToHost);
		for(int i = 0; i < cells->lattices[cell].fba.nOutputs; ++i){
			cudaMemcpy(cells->lattices[cell].fbaOuts[i], cells->lattices[cell].cFbaOuts[i], cells->N * sizeof(float), cudaMemcpyDeviceToHost);
		}
	}
}

/// @brief Copy host side cell data to device side cell data arrays.
/// @param cells Container struct of all the current cell types defined.
/// @return Void
void cellMemCopyToDev(cellLattices *cells){
	//push totVolFract to device
	cudaMemcpy(cells->cTotalVolumeFraction, cells->totalVolumeFraction, cells->N * sizeof(float), cudaMemcpyHostToDevice);
	
	//do the individual cell types:
	for(int cell = 0; cell < cells->numberCells; ++cell){
		cudaMemcpy(cells->lattices[cell].cVolumeFraction, cells->lattices[cell].volumeFraction, cells->N * sizeof(float), cudaMemcpyHostToDevice);
	}
}


//=========================================chem struct and related functions===============================
/// @brief Cell lattice struct.
/// This contains a lattice of chem concentrations, a lattice of diffusion rates, and info about time steps, boundary conditions, and other such things.
struct chemLattice{
	float *lattice;
	float *tempLattice;
	float *cLattice;
	float *cTempLattice;
	float *Dlattice;
	float *cDlattice;
	float Dwater;
	float Dagar;
	float initialAgar;
	float initialAir;
	float tau;
	int isIntracellular;
	int cellIdxChemLivesIn;
	int isHindered;
	int refWallBcFlag;
	int refFloorBcFlag;
	
// 	chemLattice(){
// 		
// 	}
	/// @brief Build a chem lattice struct.
	/// @param dw The diffusion rate in water
	/// @param da The diffusion rate in agar
	/// @param l The lattice spacing
	/// @param t The timestep
	/// @param x Number of lattice sites in the x direction
	/// @param y Number of lattice sites in the y direction
	/// @param z Number of lattice sites in the z direction
	/// @param agar The height (in lattice sites) of the agar
	/// @param initAg The initial concentration of the chemical in the agar
	/// @param initAir The initial concentration of the chemical in the air
	/// @param isInt Flag determining if the chemical species is intracellular (1) or extracellular (0)
	/// @param cicli Cell index the chemical lives is--if a chemical species is intracellular, this int denotes the index of the cell type where the chemical resides
	/// @param isHin Flag determining if a chemical experiences hindered diffusion around cells (1) of does not experience hindered diffusion (0)
	/// @param cellVol Volume of the cell (in meters^3)
	/// @param wallBcFlag Flag determining if a chemical will experience reflective boundary conditions at the 4 walls of the simulation box (1) or will experience fixed-concentration boundary conditions (0)
	/// @param floorBcFlag Flag determining if a chemical will experience reflective boundary conditions at the floor of the simulation box (1) or will experience fixed-concentration boundary conditions (0)
	/// @return Void

	void buildChemLattice(float dw, float da, float l, float t, int x, int y, int z, int agar, float initAg, float initAir, int isInt, int cicli, int isHin, float cellVol, int wallBcFlag, int floorBcFlag){
		Dwater = dw;
		Dagar = da;
		initialAgar = initAg;
		initialAir = initAir;
		isIntracellular = isInt; //change to intracellular
		cellIdxChemLivesIn = cicli;
		isHindered = isHin;
		tau = t;
		int N = (x + 2) * (y + 2) * (z + 2);
		refWallBcFlag = wallBcFlag;
		refFloorBcFlag = floorBcFlag;
		
		//malloc the basic lattices
		lattice = (float *) malloc(sizeof(float) * N);
		tempLattice = (float *) malloc(sizeof(float) * N);
		Dlattice = (float *) malloc(sizeof(float) * N);
		//now fill in the lattice:
		for(int i = 0; i < N; ++i){
			//fill in the concentration and D lattices:
			if(getZ(i, y, z) > agar){
				lattice[i] = initAir;
				tempLattice[i] = initAir;
				Dlattice[i] = dw; //give air the diffusitivity of water, because the dissolved oxygen is same as surface of water and we can keep the right AVERAGE value for D
			}
			else{
				lattice[i] = initAg;
				tempLattice[i] = initAg;
				Dlattice[i] = da;
			}
		}		
		
		//cuda mallock lattices:
		cudaMalloc((void **)&cLattice, sizeof(float) * N);
		cudaMalloc((void **)&cTempLattice, sizeof(float) * N);
		cudaMalloc((void **)&cDlattice, sizeof(float) * N);
		
		//cuda memory copy:
		cudaMemcpy(cLattice, lattice, sizeof(float) * N, cudaMemcpyHostToDevice);
		cudaMemcpy(cTempLattice, tempLattice, sizeof(float) * N, cudaMemcpyHostToDevice);
		cudaMemcpy(cDlattice, Dlattice, sizeof(float) * N, cudaMemcpyHostToDevice);
	}
};

/// @brief Chem lattices struct.
/// This is a container struct for all the chem lattice structs in the simulation.
struct chemLattices{
	// container struct for all the chem lattices; max number to track is 10.
	int xl;
	int yl;
	int zl;
	float lambda;
	float fbaTau;
	int agarHeight;
	int N;
	
	int numberChems;
	chemLattice lattices[10]; //ten max
	
	chemLattices(){
		
	}
	
	/// @brief Build the chem lattices struct
	/// @param x Number of lattice sites in the x direction
	/// @param y Number of lattice sites in the y direction
	/// @param z Number of lattice sites in the z direction
	/// @param l The lattice spacing in meters
	/// @param t The simulation timestep
	/// @param ah The agar height in lattice sites
	/// @return Void
	void buildChemLattices(int x, int y, int z, float l, float t, int ah){
		numberChems = 0;
		xl = x;
		yl = y;
		zl = z;
		lambda = l;
		fbaTau = t;
		agarHeight = ah;
		N = (xl + 2) * (yl + 2) * (zl + 2);
	}
	
	/// @brief Add a chem lattice to this container struct of chem lattices
	/// @param cl chem lattice
	/// @return Void
	void add(chemLattice cl){
		lattices[numberChems] = cl;
		numberChems++;
	}
};

/// @brief Check the input file to see if the n+1 chem lattice is described (where n is the number of chem lattices already in the chem lattices struct)
/// @param filename The input filename
/// @param chems The container struct of all the chem lattices already loaaded
/// @return True if the n+1 chem is described, false otherwise
bool isChemHere(string filename, chemLattices chems){
	ifstream inputFile;
	inputFile.open(filename.c_str());
	
	string thisLine;
	
	char buffer [4096];
	int n = sprintf(buffer, "chemType%d:\t", chems.numberChems);
	string testString = string(buffer, n);
	
	int start;
	bool eof;
	
	while(true){
		eof = (bool)((getline(inputFile, thisLine)).eof());
		if(eof){
			inputFile.close();
			return false;
		}
		start = thisLine.find(testString);
		if(start != string::npos){
			inputFile.close();
			return true;
		}
	}
}

/// @brief Get the next chem's diffusion rate in water
/// @param filename The input filename
/// @param chems The container struct of all the chem lattices already loaaded
/// @return The diffusion rate of the n+1 chem in water
float getDwater(string filename, chemLattices chems){
	char buffer [4096];
	int n = sprintf(buffer, "chemType%d dWater:\t", chems.numberChems);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((float)atof(s.c_str()));
	}
	return 0.0f;
}

/// @brief Get the next chem's diffusion rate in agar
/// @param filename The input filename
/// @param chems The container struct of all the chem lattices already loaaded
/// @return The diffusion rate of the n+1 chem in agar
float getDagar(string filename, chemLattices chems){
	char buffer [4096];
	int n = sprintf(buffer, "chemType%d dAgar:\t", chems.numberChems);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((float)atof(s.c_str()));
	}
	return 0.0f;
}

/// @brief Get the next chem's timestep
/// @param filename The input filename
/// @param chems The container struct of all the chem lattices already loaaded
/// @return The diffusion timestep for the n+1 chem
float getTau(string filename, chemLattices chems){
	char buffer [4096];
	int n = sprintf(buffer, "chemType%d tau:\t", chems.numberChems);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((float)atof(s.c_str()));
	}
	return 1e15f;
}

/// @brief Get the next chem's initial concentration in agar
/// @param filename The input filename
/// @param chems The container struct of all the chem lattices already loaded
/// @return The initial agar concentration of the next chem to be loaded
float getInitialAgarConc(string filename, chemLattices chems){
	char buffer [4096];
	int n = sprintf(buffer, "chemType%d initAgarConc:\t", chems.numberChems);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((float)atof(s.c_str()));
	}
	return 0.0f;
}

/// @brief Get the next chem's initial concentration in air
/// @param filename The input filename
/// @param chems The container struct of all the chem lattices already loaaded
/// @return The initial air concentration of the next chem to be loaded
float getInitialAirConc(string filename, chemLattices chems){
	char buffer [4096];
	int n = sprintf(buffer, "chemType%d initAirConc:\t", chems.numberChems);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((float)atof(s.c_str()));
	}
	return 0.0f;
}

/// @brief Get the next chem's intracellular flag
/// @param filename The input filename
/// @param chems The container struct of all the chem lattices already loaaded
/// @return if the next chem to be loaded is an intracellular chem, return 1, otherwise return 0
int getIntracellularFlag(string filename, chemLattices chems){
	char buffer [4096];
	int n = sprintf(buffer, "chemType%d intracellularFlag:\t", chems.numberChems);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return 0;
}

/// @brief Get the index of the cell type that the next chem exists inside, if its intracellular of course!
/// @param filename The input filename
/// @param chems The container struct of all the chem lattices already loaaded
/// @return The index of the cell type that the next chem exists inside
int getCellIdxChemLivesInFlag(string filename, chemLattices chems){
	char buffer [4096];
	int n = sprintf(buffer, "chemType%d cellIdxChemLivesIn:\t", chems.numberChems);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return -1;
}

/// @brief Get the hindered diffusion flag for the next chem to be loaded
/// @param filename The input filename
/// @param chems The container struct of all the chem lattices already loaaded
/// @return If the next chem experiences hindered diffusion, redturn 1, otherwise return 0
int getIsHinderedFlag(string filename, chemLattices chems){
	char buffer [4096];
	int n = sprintf(buffer, "chemType%d isHindered:\t", chems.numberChems);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return 0;
}

/// @brief Get the reflective wall boundary condition flag
/// @param filename The input filename
/// @param chems The container struct of all the chem lattices already loaaded
/// @return If the next chem experiences reflective boundary conditions on the walls of the simulation box return 1, otherwise 0
int getRefWallBcFlag(string filename, chemLattices chems){
	char buffer [4096];
	int n = sprintf(buffer, "chemType%d refWallBcFlag:\t", chems.numberChems);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return 0;
}

/// @brief Get the reflective floor boundary condition flag
/// @param filename The input filename
/// @param chems The container struct of all the chem lattices already loaaded
/// @return If the next chem experiences reflective boundary conditions on the floor of the simulation box return 1, otherwise 0
int getRefFloorBcFlag(string filename, chemLattices chems){
	char buffer [4096];
	int n = sprintf(buffer, "chemType%d refFloorBcFlag:\t", chems.numberChems);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return 0;
}

/// @brief swap pointers between the chem lattices and the temporary chem lattices on the host
/// @param chems The container struct of all the chem lattices
/// @return Void
void swapChemLattices_host(chemLattices *chems){
	for(int i = 0; i < chems->numberChems; ++i){
		float *temp = chems->lattices[i].cLattice;
		chems->lattices[i].cTempLattice = chems->lattices[i].cLattice;
		chems->lattices[i].cLattice = temp;
	}
}

/// @brief swap pointers between A chem lattice and its corresponding temporary chem lattice on the host
/// @param chems The container struct of all the chem lattices
/// @param chemIdx The index of the chem whose pointers should be swapped
/// @return Void
void swapAChemLattice_host(chemLattices *chems, int chemIdx){
	float *temp = chems->lattices[chemIdx].cLattice;
	chems->lattices[chemIdx].cTempLattice = chems->lattices[chemIdx].cLattice;
	chems->lattices[chemIdx].cLattice = temp;
}

/// @brief copy all the chem lattices back to host
/// @param chems The container struct of all the chem lattices
/// @return Void
void chemMemCopyBack(chemLattices *chems){
	//for each chem bring back the lattice:
	for(int chem = 0; chem < chems->numberChems; ++chem){
		cudaMemcpy(chems->lattices[chem].lattice, chems->lattices[chem].cLattice, chems->N * sizeof(float), cudaMemcpyDeviceToHost);
	}
}

/// @brief copy all the chem lattices to the GPU
/// @param chems The container struct of all the chem lattices
/// @return Void
void chemMemCopyToDev(chemLattices *chems){
	//for each chem push it to device:
	for(int chem = 0; chem < chems->numberChems; ++chems){
		cudaMemcpy(chems->lattices[chem].cLattice, chems->lattices[chem].lattice, chems->N * sizeof(float), cudaMemcpyHostToDevice);
	}
}


//=====================================reactions structs and related functions====================================================
/// @brief Chem uptake reaction struct
struct chemReaction{
	
	float km;
	float kcat;
	float E; //density of enzymes in a lambda^3 subvolume of full cell packing
	int cellType;
	int reactant;
	int product;
	
	chemReaction(){
		
	}
	
	/// @brief Build an uptake reaction struct
	/// @param m Michaelis constant for a michaelis-menten uptake reaction
	/// @param cat Kcat for a michaelis-menten uptake reaction
	/// @param e Number of uptake enzymes on an average cell
	/// @param type The index of the cell type that does this uptake reaction
	/// @param react The index of the chem type that is uptaken
	/// @param prod The index of the chem type that is produced (intercellular)
	/// @param cellV The volume of the cell type
	void buildChemReaction(float m, float cat, float e, int type, int react, int prod, float cellV){
		km = m;
		kcat = cat;
		cellType = type;
		reactant = react;
		product = prod;
		E = e / ((cellV * 1000.0f) * 6.02e23); //e is the number of trnsporters on a cell.
	}
	
};

/// @brief container struct for all the uptake reaction structs
struct chemReactions{
	
	int numberReactions;
	chemReaction reactions[10]; //max ten reactions
	
	chemReactions(){
		numberReactions = 0;
	}
	
	/// @brief add an uptake reaction to the container struct
	/// @param cR The uptake reaction struct to be added
	/// @return Void
	void add(chemReaction cR){
		reactions[numberReactions] = cR;
		numberReactions++;
	}
};

/// @brief Check if the input file contains the an n+1 reaction, given n reactions already in the container struct
/// @param filename The name of the input file
/// @param rxns The container struct for the uptake reactions already loaded
/// @return True if the n+1 reaction is described in the input file, false otherwise
bool isRxnHere(string filename, chemReactions rxns){
	ifstream inputFile;
	inputFile.open(filename.c_str());
	
	string thisLine;
	
	char buffer [4096];
	int n = sprintf(buffer, "rxnType%d:\t", rxns.numberReactions);
	string testString = string(buffer, n);
	
	int start;
	bool eof;
	
	while(true){
		eof = (bool)((getline(inputFile, thisLine)).eof());
		if(eof){
			inputFile.close();
			return false;
		}
		start = thisLine.find(testString);
		if(start != string::npos){
			inputFile.close();
			return true;
		}
	}
}

/// @brief Get the michaelis constant for the n+1 uptake reaction
/// @param filename The name of the input file
/// @param rxns The container struct for the uptake reactions already loaded
/// @return The Michaelis constant for the n+1 uptake reaction
float getKm(string filename, chemReactions rxns){
	char buffer [4096];
	int n = sprintf(buffer, "rxnType%d km:\t", rxns.numberReactions);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((float)atof(s.c_str()));
	}
	return -1.0;
}

/// @brief Get the Kcat for the n+1 uptake reaction
/// @param filename The name of the input file
/// @param rxns The container struct for the uptake reactions already loaded
/// @return The Kcat parameter for the n+1 uptake reaction
float getKcat(string filename, chemReactions rxns){
	char buffer [4096];
	int n = sprintf(buffer, "rxnType%d kcat:\t", rxns.numberReactions);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((float)atof(s.c_str()));
	}
	return -1.0;
}

/// @brief Get the average number of transport enzymes per cell for the n+1 uptake reaction
/// @param filename The name of the input file
/// @param rxns The container struct for the uptake reactions already loaded
/// @return The average number of transport enzymes per cell for the n+1 uptake reaction
float getEnzymeNumberPerCell(string filename, chemReactions rxns){
	char buffer [4096];
	int n = sprintf(buffer, "rxnType%d E:\t", rxns.numberReactions);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((float)atof(s.c_str()));
	}
	return -1.0;
}

/// @brief Get the cell type index for the n+1 uptake reaction
/// @param filename The name of the input file
/// @param rxns The container struct for the uptake reactions already loaded
/// @return The index of the cell type that does the n+1 uptake reaction
int getCellTypeThatDoesThisReaction(string filename, chemReactions rxns){
	char buffer [4096];
	int n = sprintf(buffer, "rxnType%d cellType:\t", rxns.numberReactions);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return -1;
}

/// @brief Get the reactant chem type index for the n+1 uptake reaction
/// @param filename The name of the input file
/// @param rxns The container struct for the uptake reactions already loaded
/// @return The index of the reactant chem type for the n+1 uptake reaction
int getReactantChemIndex(string filename, chemReactions rxns){
	char buffer [4096];
	int n = sprintf(buffer, "rxnType%d reactantIndex:\t", rxns.numberReactions);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return -1;
}

/// @brief Get the product chem type index for the n+1 uptake reaction
/// @param filename The name of the input file
/// @param rxns The container struct for the uptake reactions already loaded
/// @return The index of the product chem type for the n+1 uptake reaction (should be intracellular type)
int getProductChemIndex(string filename, chemReactions rxns){
	char buffer [4096];
	int n = sprintf(buffer, "rxnType%d productIndex:\t", rxns.numberReactions);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return -1;
}


//===========================================regulation struct and related functions===================================
/// @brief Chem cognate struct
struct chemCognate{ //chem internal to one cell type may have an identical chem in another cell type
	
	int reactChemIdx;
	int prodChemIdx;
	
	chemCognate(){
	
	}
};

/// @brief Regulation reaction struct
struct regReaction{
	
	float multipliers[10];
	chemCognate cognates[10];
	int numChemCognates;
	int chemType1;
	int chemType2;
	int reactantCell;
	int productCell;
	
	regReaction(){
		
	}
	
	/// @brief Build a regulation reaction
	/// @param chem1 Index of the first chemical species a cell is assumed to change regulatory state in response to
	/// @param chem2 Index of the second chemical species a cell is assumed to change regulatory state in response to
	/// @param reactCell Index of the cell type transformed by this regulation reaction
	/// @param prodCell Index of the cell type produced by this regulation reaction
	/// @param mults Coefficients for the polynomial rate law that governs this reaction
	/// @return Void
	void buildRegReaction(int chem1, int chem2, int reactCell, int prodCell, string mults){
		chemType1 = chem1;
		chemType2 = chem2;
		reactantCell = reactCell;
		productCell = prodCell;
		numChemCognates = 0;
		
		vector< string > multsV = split(mults, "\t");
		for(int i = 0; i < 10; ++i){
			multipliers[i] = (float)atof(multsV[i].c_str());
		}
	}
	
	/// @brief Regulation reactions turn one cell type into another; the intracellular species within the reactant cell type also need to transform into the intracellular species of the product cell type
	/// @param cogs Tab delimited set of indexes of the intracellular chem index in the reactant cell (first) and product cell (second)
	/// @return Void
	void fillChemCognate(string cogs){
	
		vector< string > cogsV = split(cogs, "\t");

		cognates[numChemCognates].reactChemIdx = (int)atoi(cogsV[0].c_str());
		cognates[numChemCognates].prodChemIdx = (int)atoi(cogsV[1].c_str());
		numChemCognates++;
	}
};

/// @brief container struct for the regulation reactions
struct regReactions{
	
	int numberReactions;
	regReaction reactions[10]; //max ten reactions
	
	regReactions(){
		numberReactions = 0;
	}
	
	/// @brief add a regulation reaction to this container struct
	/// @param reg The regulation reaction struct to be added
	/// @return Void
	void add(regReaction reg){
		reactions[numberReactions] = reg;
		numberReactions++;
	}
};

/// @brief Check if the input file contains the an n+1 regulation reaction, given n regulation reactions are already in the container struct
/// @param filename The name of the input file
/// @param regs The container struct for the regulation reactions already loaded
/// @return True if the n+1 regulation reaction is described in the input file, false otherwise
bool isRegHere(string filename, regReactions regs){
	ifstream inputFile;
	inputFile.open(filename.c_str());
	
	string thisLine;
	
	char buffer [4096];
	int n = sprintf(buffer, "regType%d:\t", regs.numberReactions);
	string testString = string(buffer, n);
	
	int start;
	bool eof;
	
	while(true){
		eof = (bool)((getline(inputFile, thisLine)).eof());
		if(eof){
			inputFile.close();
			return false;
		}
		start = thisLine.find(testString);
		if(start != string::npos){
			inputFile.close();
			return true;
		}
	}
}

/// @brief Check if the input file contains the an n+1 set of chemical cognates for regulation reaction m, given n cognates are already loaded for regulation reaction m
/// @param filename The name of the input file
/// @param regs The container struct for the regulation reactions already loaded
/// @param reg The regulation reaction currently being built
/// @return True if the n+1 cognate for the mth regulation reaction is described in the input file, false otherwise
bool isCogHere(string filename, regReactions regs, regReaction reg){
	ifstream inputFile;
	inputFile.open(filename.c_str());
	
	string thisLine;
	
	char buffer [4096];
	int n = sprintf(buffer, "regType%d chemCognate%d:\t", regs.numberReactions, reg.numChemCognates);
	string testString = string(buffer, n);
	
	int start;
	bool eof;
	
	while(true){
		eof = (bool)((getline(inputFile, thisLine)).eof());
		if(eof){
			inputFile.close();
			return false;
		}
		start = thisLine.find(testString);
		if(start != string::npos){
			inputFile.close();
			return true;
		}
	}
}

/// @brief Get the chem1 index of the n+1 regulation reaction, given n regulation reactions have laready been loaded
/// @param filename The name of the input file
/// @param regs The container struct for the regulation reactions already loaded
/// @return Index of the chem1
int getRegChemOne(string filename, regReactions regs){
	char buffer [4096];
	int n = sprintf(buffer, "regType%d chemOneIndex:\t", regs.numberReactions);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return -1;
}

/// @brief Get the chem2 index of the n+1 regulation reaction, given n regulation reactions have laready been loaded
/// @param filename The name of the input file
/// @param regs The container struct for the regulation reactions already loaded
/// @return Index of the chem2
int getRegChemTwo(string filename, regReactions regs){
	char buffer [4096];
	int n = sprintf(buffer, "regType%d chemTwoIndex:\t", regs.numberReactions);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return -1;
}

/// @brief Get the index of reactant cell type for the n+1 regulation reaction, given n regulation reactions have laready been loaded
/// @param filename The name of the input file
/// @param regs The container struct for the regulation reactions already loaded
/// @return Index of the reactant cell type
int getRegReactCell(string filename, regReactions regs){
	char buffer [4096];
	int n = sprintf(buffer, "regType%d reactantCellType:\t", regs.numberReactions);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return -1;
}

/// @brief Get the index of product cell type for the n+1 regulation reaction, given n regulation reactions have laready been loaded
/// @param filename The name of the input file
/// @param regs The container struct for the regulation reactions already loaded
/// @return Index of the product cell type
int getRegProdCell(string filename, regReactions regs){
	char buffer [4096];
	int n = sprintf(buffer, "regType%d productCellType:\t", regs.numberReactions);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	if(s.length() > 0){
		return ((int)atoi(s.c_str()));
	}
	return -1;
}

/// @brief Get the polynomial coefficients for the regulation reaction rate
/// @param filename The name of the input file
/// @param regs The container struct for the regulation reactions already loaded
/// @return String of tab-delimited polynomial coefficients
string getRegMultipliersString(string filename, regReactions regs){
	char buffer [4096];
	int n = sprintf(buffer, "regType%d multipliers:\t", regs.numberReactions);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	return s;
}

/// @brief Get the data on the n+1 set of chem cognates for regulation reaction m
/// @param filename The name of the input file
/// @param regs The container struct for the regulation reactions already loaded
/// @param reg The regulation reaction currently being built
/// @return tab-delimited string of cognate chem indexes
string getChemCognate(string filename, regReactions regs, regReaction reg){
	char buffer [4096];
	int n = sprintf(buffer, "regType%d chemCognate%d:\t", regs.numberReactions, reg.numChemCognates);
	string testString = string(buffer, n);
	
	string s = getStringAfterIdentifier(filename, testString);
	return s;
}

/// @brief struct of block dimensions for CUDA kernels
struct blockDims{
	dim3 defaultNumBlocks;
	dim3 fbaNumBlocks;
	dim3 diffNumBlocks;
	dim3 syncTCLwithCLNumBlocks;
	dim3 chemReactNumBlocks;
	dim3 expandCellsNumBlocks;
	
	int BCblocksWall;
	int BCblocksFloor;
	
	blockDims(){
	
	}
	
	/// @brief make up struct of block dimentions for cuda kernels
	/// @param xl lattice length in x dimensions
	/// @param yl lattice length in y dimensions
	/// @param zl lattice length in z dimensions
	/// @return Void
	void makeBlockDims(int xl, int yl, int zl){
		int n = xl * yl * zl;
		int thisNumBlocks;
		
		thisNumBlocks = n / defaultThreads;
		dim3 defaultBlockDim(32, thisNumBlocks / 32);
		defaultNumBlocks = defaultBlockDim;
		
		thisNumBlocks = n / fbaThreads;
		dim3 fbaBlockDim(32, thisNumBlocks / 32);
		fbaNumBlocks = fbaBlockDim;
		
		thisNumBlocks = n / diffThreads;
		dim3 diffBlockDim(32, thisNumBlocks / 32);
		diffNumBlocks = diffBlockDim;
		
		thisNumBlocks = n / syncTCLwithCLThreads;
		dim3 syncBlockDim(32, thisNumBlocks / 32);
		syncTCLwithCLNumBlocks = syncBlockDim;
		
		thisNumBlocks = n / chemReactThreads;
		dim3 chemRBlockDim(32, thisNumBlocks / 32);
		chemReactNumBlocks = chemRBlockDim;
		
		thisNumBlocks = n / expandCellsThreads;
		dim3 expandBlockDim(32, thisNumBlocks / 32);
		expandCellsNumBlocks = expandBlockDim;
		
		BCblocksWall = max(
							((yl * zl) % defaultThreads == 0) ? ((yl * zl) / defaultThreads) : ((yl * zl) / defaultThreads + 1),
							((xl * zl) % defaultThreads == 0) ? ((xl * zl) / defaultThreads) : ((xl * zl) / defaultThreads + 1)
						);
		BCblocksFloor = ((xl * yl) % defaultThreads == 0) ? ((xl * yl) / defaultThreads) : ((xl * yl) / defaultThreads + 1);
	}
	
};

//===================================================IO FUNCTIONS===================================================================

/// @brief open existing .tar.gx file for restart
/// @param filename
/// @return Void
void openExistingZippedDirectory(string filename){
	
	string fn = filename.substr(0, filename.find("."));
	
	string copyCommand = "cp " + fn + ".tar.gz " + fn + "_tmp.tar.gz";
	system(copyCommand.c_str());
	
	string tarCommand = "tar -zxf " + fn + "_tmp.tar.gz";
	system(tarCommand.c_str());
	
}

/// @brief compress output directory as .tar.gz
/// @param filename
/// @return Void
void closeAndZipDirectory(string filename){
	
	string fn = filename.substr(0, filename.find("."));
	
	string tarCommand = "tar -zcf " + fn + "_tmp.tar.gz " + fn;
	system(tarCommand.c_str());
	
	string moveCommand = "mv " + fn + "_tmp.tar.gz " + fn + ".tar.gz";
	system(moveCommand.c_str());
	
}

/// @brief create a new directory tree for simulation output
/// @param filename
/// @param cells struct containing al the cell lattices
/// @param chems struct containing all the chem lattices
/// @return Void
void creatNewOutputDirectory(string filename, cellLattices *cells, chemLattices *chems){
	string fn = filename.substr(0, filename.find("."));
	
	string createCommand = "mkdir " + fn;
	system(createCommand.c_str());
	
	string copyInpCommand = "cp " + filename + " " + fn + "/";
	system(copyInpCommand.c_str());
	
	system((createCommand + "/chems").c_str());
	system((createCommand + "/cells").c_str());
	system((createCommand + "/cells/totalVolumeFraction").c_str());
	
	char buffer [4096];
	//do the cell stuff
	for(int cell = 0; cell < cells->numberCells; ++cell){
		
		int n = sprintf(buffer, "/cells/%d", cell);
		string thisCellAddendum = string(buffer, n);
		system((createCommand + thisCellAddendum).c_str());
		system((createCommand + thisCellAddendum + "/growthRate").c_str());
		system((createCommand + thisCellAddendum + "/volumeFraction").c_str());
		
		//do the fluxes:
		for(int fbaInp = 0; fbaInp < cells->lattices[cell].fba.nOutputs - 1; ++fbaInp){
			
			int m = sprintf(buffer, "/chem%dflux", cells->lattices[cell].fba.outputIndexes[fbaInp]);
			string thisChemFluxAddendum = string(buffer, m);
			system((createCommand + thisCellAddendum + thisChemFluxAddendum).c_str());
		}
	}
	
	//do the chem stuff
	for(int chem = 0; chem < chems->numberChems; ++chem){
		
		int n = sprintf(buffer, "/chems/%d", chem);
		string thisChemAddendum = string(buffer, n);
		system((createCommand + thisChemAddendum).c_str());
		
	}
}


/// @brief write binary lattice data for every chem and cell type at a given timepoint
/// @param filename
/// @param cells struct containing al the cell lattices
/// @param chems struct containing all the chem lattices
/// @param time
/// @return Void
void writeToExisting(string filename, cellLattices *cells, chemLattices *chems, float time){
	string fn = filename.substr(0, filename.find("."));
	ofstream thisOutputStream;
	char buffer [4096];
	int n;
	
	//do the cell stuff
	//starting with the totalVolumeFraction
	n = sprintf(buffer, "/cells/totalVolumeFraction/%.6E", time);
	string thisAddendum = string(buffer, n);
	thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::trunc | ios_base::binary);
	thisOutputStream.write((char *)cells->totalVolumeFraction, sizeof(float) * cells->N);
	thisOutputStream.close();
	
	for(int cell = 0; cell < cells->numberCells; ++cell){
		
		//growthRate first
		n = sprintf(buffer, "/cells/%d/growthRate/%.6E", cell, time);
		thisAddendum = string(buffer, n);
		
		thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::trunc | ios_base::binary);
		thisOutputStream.write((char *)cells->lattices[cell].fbaOuts[cells->lattices[cell].fba.nOutputs - 1], sizeof(float) * cells->N);
		thisOutputStream.close();
		
		//volumeFraction second
		n = sprintf(buffer, "/cells/%d/volumeFraction/%.6E", cell, time);
		thisAddendum = string(buffer, n);
		
		thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::trunc | ios_base::binary);
		thisOutputStream.write((char *)cells->lattices[cell].volumeFraction, sizeof(float) * cells->N);
		thisOutputStream.close();

		//do the fluxes:
		for(int fbaInp = 0; fbaInp < cells->lattices[cell].fba.nOutputs - 1; ++fbaInp){
			n = sprintf(buffer, "/cells/%d/chem%dflux/%.6E", cell, cells->lattices[cell].fba.outputIndexes[fbaInp], time);
			thisAddendum = string(buffer, n);
			
			thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::trunc | ios_base::binary);
			thisOutputStream.write((char *)cells->lattices[cell].fbaOuts[fbaInp], sizeof(float) * cells->N);
			thisOutputStream.close();
		}
	}
	
	//chems
	for(int chem = 0; chem < chems->numberChems; ++chem){
		
		n = sprintf(buffer, "/chems/%d/%.6E", chem, time);
		thisAddendum = string(buffer, n);
		
		thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::trunc | ios_base::binary);
		thisOutputStream.write((char *)chems->lattices[chem].lattice, sizeof(float) * chems->N);
		thisOutputStream.close();
		
	}
	
	//time last
	thisOutputStream.open((fn + "/time").c_str(), ios_base::trunc | ios_base::binary);
	thisOutputStream.write((char *)&time, sizeof(float));
	thisOutputStream.close();
	
}

/// @brief load chem and cell data from an existing file, as a restart.
/// @param filename
/// @param cells struct containing al the cell lattices
/// @param chems struct containing all the chem lattices
/// @return Void
float restartFromExisting(string filename, cellLattices *cells, chemLattices *chems){
	string fn = filename.substr(0, filename.find("."));
	ifstream thisInputStream;
	char buffer [4096];
	int n;
	
	float t;
	
	//do the time first:
	thisInputStream.open((fn + "/time").c_str(), ios_base::binary);
	thisInputStream.read((char *)&t, sizeof(float));
	thisInputStream.close();
	
	//do everything else
	//do the cell stuff
	//starting with the totalVolumeFraction
	n = sprintf(buffer, "/cells/totalVolumeFraction/%.6E", t);
	string thisAddendum = string(buffer, n);
	thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
	thisInputStream.read((char *)cells->totalVolumeFraction, sizeof(float) * cells->N);
	thisInputStream.close();
	
	for(int cell = 0; cell < cells->numberCells; ++cell){
		
		//growthRate first
		n = sprintf(buffer, "/cells/%d/growthRate/%.6E", cell, t);
		thisAddendum = string(buffer, n);
		
		thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
		thisInputStream.read((char *)cells->lattices[cell].fbaOuts[cells->lattices[cell].fba.nOutputs - 1], sizeof(float) * cells->N);
		thisInputStream.close();
		
		//volumeFraction second
		n = sprintf(buffer, "/cells/%d/volumeFraction/%.6E", cell, t);
		thisAddendum = string(buffer, n);
		
		thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
		thisInputStream.read((char *)cells->lattices[cell].volumeFraction, sizeof(float) * cells->N);
		thisInputStream.close();

		//do the fluxes:
		for(int fbaInp = 0; fbaInp < cells->lattices[cell].fba.nOutputs - 1; ++fbaInp){
			n = sprintf(buffer, "/cells/%d/chem%dflux/%.6E", cell, cells->lattices[cell].fba.outputIndexes[fbaInp], t);
			thisAddendum = string(buffer, n);
			
			thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
			thisInputStream.read((char *)cells->lattices[cell].fbaOuts[fbaInp], sizeof(float) * cells->N);
			thisInputStream.close();
		}
	}
	
	//chems
	for(int chem = 0; chem < chems->numberChems; ++chem){
		
		n = sprintf(buffer, "/chems/%d/%.6E", chem, t);
		thisAddendum = string(buffer, n);
		
		thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
		thisInputStream.read((char *)chems->lattices[chem].lattice, sizeof(float) * chems->N);
		thisInputStream.close();
		
	}
	
	//return the time:
	return t;
}

void writeYZSlice(){
	
	
	
}

void writeXZSlice(){
	
	
	
}

void writeXYSlice(){
	
	
	
}

/// @brief write a trajectory of every chem and cell type in the simulation to plane txt files at a given YZ plane
/// @param filename
/// @param cells struct containing al the cell lattices
/// @param chems struct containing all the chem lattices
/// @param Xslice x=Xslice is the YZ plane written out
/// @param growT the time between growth events
/// @param pullbackSteps the time interval at which the simulation state was saved
/// @return Void
void writeYZTrajectory(string filename, cellLattices *cells, chemLattices *chems, int Xslice, float growT, int pullbackSteps){
	string fn = filename.substr(0, filename.find("."));
	ifstream thisInputStream;
	ofstream thisOutputStream;
	char buffer [4096];
	int n;
	
	//get xl, yl, zl, and time
	//int xl = cells->xl;
	int yl = cells->yl;
	int zl = cells->zl;
	int N = cells->N;
	float t;
	
	//do the time first:
	thisInputStream.open((fn + "/time").c_str(), ios_base::binary);
	thisInputStream.read((char *)&t, sizeof(float));
	thisInputStream.close();
	
	
	float *lattice;
	lattice = (float *) malloc(sizeof(float) * N);
	string thisAddendum;
	
	//loop over times:
	float thisT = 0.0f;
	while(!(thisT > t)){
		
		//do the cell stuff
		//starting with the totalVolumeFraction
		n = sprintf(buffer, "/cells/totalVolumeFraction/%.6E", thisT);
		thisAddendum = string(buffer, n);
		thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
		thisInputStream.read((char *)lattice, sizeof(float) * N);
		thisInputStream.close();
		
		n = sprintf(buffer, "_totalVolumeFraction_x%d.traj", Xslice);
		thisAddendum = string(buffer, n);
		thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
		thisOutputStream << "\nt = " << thisT << "\n";
		for(int k = 0; k < zl; ++k){
			for(int j = 0; j < yl; ++j){
				thisOutputStream << lattice[idx(Xslice, j, (zl - k - 1), yl, zl)] << "\t";
			}
			thisOutputStream << "\n";
		}
		thisOutputStream.close();
		
		//individual cells now
		for(int cell = 0; cell < cells->numberCells; ++cell){
			
			//growthRate first
			n = sprintf(buffer, "/cells/%d/growthRate/%.6E", cell, thisT);
			thisAddendum = string(buffer, n);
			
			thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
			thisInputStream.read((char *)lattice, sizeof(float) * N);
			thisInputStream.close();
			
			n = sprintf(buffer, "_cell%d_growthRate_x%d.traj", cell, Xslice);
			thisAddendum = string(buffer, n);
			thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
			thisOutputStream << "\nt = " << thisT << "\n";
			for(int k = 0; k < zl; ++k){
				for(int j = 0; j < yl; ++j){
					thisOutputStream << lattice[idx(Xslice, j, (zl - k - 1), yl, zl)] << "\t";
				}
				thisOutputStream << "\n";
			}
			thisOutputStream.close();
			
			
			//volumeFraction second
			n = sprintf(buffer, "/cells/%d/volumeFraction/%.6E", cell, thisT);
			thisAddendum = string(buffer, n);
			
			thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
			thisInputStream.read((char *)lattice, sizeof(float) * N);
			thisInputStream.close();
			
			n = sprintf(buffer, "_cell%d_volumeFraction_x%d.traj", cell, Xslice);
			thisAddendum = string(buffer, n);
			thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
			thisOutputStream << "\nt = " << thisT << "\n";
			for(int k = 0; k < zl; ++k){
				for(int j = 0; j < yl; ++j){
					thisOutputStream << lattice[idx(Xslice, j, (zl - k - 1), yl, zl)] << "\t";
				}
				thisOutputStream << "\n";
			}
			thisOutputStream.close();
			
			//do the fluxes:
			for(int fbaInp = 0; fbaInp < cells->lattices[cell].fba.nOutputs - 1; ++fbaInp){
				n = sprintf(buffer, "/cells/%d/chem%dflux/%.6E", cell, cells->lattices[cell].fba.outputIndexes[fbaInp], thisT);
				thisAddendum = string(buffer, n);
				
				thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
				thisInputStream.read((char *)lattice, sizeof(float) * N);
				thisInputStream.close();
				
				n = sprintf(buffer, "_cell%d_chem%dflux_x%d.traj", cell, cells->lattices[cell].fba.outputIndexes[fbaInp], Xslice);
				thisAddendum = string(buffer, n);
				thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
				thisOutputStream << "\nt = " << thisT << "\n";
				for(int k = 0; k < zl; ++k){
					for(int j = 0; j < yl; ++j){
						thisOutputStream << lattice[idx(Xslice, j, (zl - k - 1), yl, zl)] << "\t";
					}
					thisOutputStream << "\n";
				}
				thisOutputStream.close();
			}
		}
		
		//chems
		for(int chem = 0; chem < chems->numberChems; ++chem){
			
			n = sprintf(buffer, "/chems/%d/%.6E", chem, thisT);
			thisAddendum = string(buffer, n);
			
			thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
			thisInputStream.read((char *)lattice, sizeof(float) * N);
			thisInputStream.close();
			
			n = sprintf(buffer, "_chem%d_x%d.traj", chem, Xslice);
			thisAddendum = string(buffer, n);
			thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
			thisOutputStream << "\nt = " << thisT << "\n";
			for(int k = 0; k < zl; ++k){
				for(int j = 0; j < yl; ++j){
					thisOutputStream << lattice[idx(Xslice, j, (zl - k - 1), yl, zl)] << "\t";
				}
				thisOutputStream << "\n";
			}
			thisOutputStream.close();
		}
		
		thisT += growT * pullbackSteps;
	}
}

/// @brief write a trajectory of every chem and cell type in the simulation to plane txt files at a given XZ plane
/// @param filename
/// @param cells struct containing al the cell lattices
/// @param chems struct containing all the chem lattices
/// @param Yslice y=Yslice is the XZ plane written out
/// @param growT the time between growth events
/// @param pullbackSteps the time interval at which the simulation state was saved
/// @return Void
void writeXZTrajectory(string filename, cellLattices *cells, chemLattices *chems, int Yslice, float growT, int pullbackSteps){
	string fn = filename.substr(0, filename.find("."));
	ifstream thisInputStream;
	ofstream thisOutputStream;
	char buffer [4096];
	int n;
	
	//get xl, yl, zl, and time
	int xl = cells->xl;
	int yl = cells->yl;
	int zl = cells->zl;
	int N = cells->N;
	float t;
	
	//do the time first:
	thisInputStream.open((fn + "/time").c_str(), ios_base::binary);
	thisInputStream.read((char *)&t, sizeof(float));
	thisInputStream.close();
	
	
	float *lattice;
	lattice = (float *) malloc(sizeof(float) * N);
	string thisAddendum;
	
	//loop over times:
	float thisT = 0.0f;
	while(!(thisT > t)){
		
		//do the cell stuff
		//starting with the totalVolumeFraction
		n = sprintf(buffer, "/cells/totalVolumeFraction/%.6E", thisT);
		thisAddendum = string(buffer, n);
		thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
		thisInputStream.read((char *)lattice, sizeof(float) * N);
		thisInputStream.close();
		
		n = sprintf(buffer, "_totalVolumeFraction_y%d.traj", Yslice);
		thisAddendum = string(buffer, n);
		thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
		thisOutputStream << "\nt = " << thisT << "\n";
		for(int k = 0; k < zl; ++k){
			for(int i = 0; i < xl; ++i){
				thisOutputStream << lattice[idx(i, Yslice, (zl - k - 1), yl, zl)] << "\t";
			}
			thisOutputStream << "\n";
		}
		thisOutputStream.close();
		
		//individual cells now
		for(int cell = 0; cell < cells->numberCells; ++cell){
			
			//growthRate first
			n = sprintf(buffer, "/cells/%d/growthRate/%.6E", cell, thisT);
			thisAddendum = string(buffer, n);
			
			thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
			thisInputStream.read((char *)lattice, sizeof(float) * N);
			thisInputStream.close();
			
			n = sprintf(buffer, "_cell%d_growthRate_y%d.traj", cell, Yslice);
			thisAddendum = string(buffer, n);
			thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
			thisOutputStream << "\nt = " << thisT << "\n";
			for(int k = 0; k < zl; ++k){
				for(int i = 0; i < xl; ++i){
					thisOutputStream << lattice[idx(i, Yslice, (zl - k - 1), yl, zl)] << "\t";
				}
				thisOutputStream << "\n";
			}
			thisOutputStream.close();
			
			
			//volumeFraction second
			n = sprintf(buffer, "/cells/%d/volumeFraction/%.6E", cell, thisT);
			thisAddendum = string(buffer, n);
			
			thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
			thisInputStream.read((char *)lattice, sizeof(float) * N);
			thisInputStream.close();
			
			n = sprintf(buffer, "_cell%d_volumeFraction_y%d.traj", cell, Yslice);
			thisAddendum = string(buffer, n);
			thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
			thisOutputStream << "\nt = " << thisT << "\n";
			for(int k = 0; k < zl; ++k){
				for(int i = 0; i < xl; ++i){
					thisOutputStream << lattice[idx(i, Yslice, (zl - k - 1), yl, zl)] << "\t";
				}
				thisOutputStream << "\n";
			}
			thisOutputStream.close();
			
			//do the fluxes:
			for(int fbaInp = 0; fbaInp < cells->lattices[cell].fba.nOutputs - 1; ++fbaInp){
				n = sprintf(buffer, "/cells/%d/chem%dflux/%.6E", cell, cells->lattices[cell].fba.outputIndexes[fbaInp], thisT);
				thisAddendum = string(buffer, n);
				
				thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
				thisInputStream.read((char *)lattice, sizeof(float) * N);
				thisInputStream.close();
				
				n = sprintf(buffer, "_cell%d_chem%dflux_y%d.traj", cell, cells->lattices[cell].fba.outputIndexes[fbaInp], Yslice);
				thisAddendum = string(buffer, n);
				thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
				thisOutputStream << "\nt = " << thisT << "\n";
				for(int k = 0; k < zl; ++k){
					for(int i = 0; i < xl; ++i){
						thisOutputStream << lattice[idx(i, Yslice, (zl - k - 1), yl, zl)] << "\t";
					}
					thisOutputStream << "\n";
				}
				thisOutputStream.close();
			}
		}
		
		//chems
		for(int chem = 0; chem < chems->numberChems; ++chem){
			
			n = sprintf(buffer, "/chems/%d/%.6E", chem, thisT);
			thisAddendum = string(buffer, n);
			
			thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
			thisInputStream.read((char *)lattice, sizeof(float) * N);
			thisInputStream.close();
			
			n = sprintf(buffer, "_chem%d_y%d.traj", chem, Yslice);
			thisAddendum = string(buffer, n);
			thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
			thisOutputStream << "\nt = " << thisT << "\n";
			for(int k = 0; k < zl; ++k){
				for(int i = 0; i < xl; ++i){
					thisOutputStream << lattice[idx(i, Yslice, (zl - k - 1), yl, zl)] << "\t";
				}
				thisOutputStream << "\n";
			}
			thisOutputStream.close();
		}
		
		thisT += growT * pullbackSteps;
	}
}


/// @brief write a trajectory of every chem and cell type in the simulation to plane txt files at a given XY plane
/// @param filename
/// @param cells struct containing al the cell lattices
/// @param chems struct containing all the chem lattices
/// @param Zslice z=Zslice is the XY plane written out
/// @param growT the time between growth events
/// @param pullbackSteps the time interval at which the simulation state was saved
/// @return Void
void writeXYTrajectory(string filename, cellLattices *cells, chemLattices *chems, int Zslice, float growT, int pullbackSteps){
	string fn = filename.substr(0, filename.find("."));
	ifstream thisInputStream;
	ofstream thisOutputStream;
	char buffer [4096];
	int n;
	
	//get xl, yl, zl, and time
	int xl = cells->xl;
	int yl = cells->yl;
	int zl = cells->zl;
	int N = cells->N;
	float t;
	
	//do the time first:
	thisInputStream.open((fn + "/time").c_str(), ios_base::binary);
	thisInputStream.read((char *)&t, sizeof(float));
	thisInputStream.close();
	
	
	float *lattice;
	lattice = (float *) malloc(sizeof(float) * N);
	string thisAddendum;
	
	//loop over times:
	float thisT = 0.0f;
	while(!(thisT > t)){
		
		//do the cell stuff
		//starting with the totalVolumeFraction
		n = sprintf(buffer, "/cells/totalVolumeFraction/%.6E", thisT);
		thisAddendum = string(buffer, n);
		thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
		thisInputStream.read((char *)lattice, sizeof(float) * N);
		thisInputStream.close();
		
		n = sprintf(buffer, "_totalVolumeFraction_z%d.traj", Zslice);
		thisAddendum = string(buffer, n);
		thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
		thisOutputStream << "\nt = " << thisT << "\n";
		for(int j = 0; j < yl; ++j){
			for(int i = 0; i < xl; ++i){
				thisOutputStream << lattice[idx(i, (yl - j - 1), Zslice, yl, zl)] << "\t";
			}
			thisOutputStream << "\n";
		}
		thisOutputStream.close();
		
		//individual cells now
		for(int cell = 0; cell < cells->numberCells; ++cell){
			
			//growthRate first
			n = sprintf(buffer, "/cells/%d/growthRate/%.6E", cell, thisT);
			thisAddendum = string(buffer, n);
			
			thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
			thisInputStream.read((char *)lattice, sizeof(float) * N);
			thisInputStream.close();
			
			n = sprintf(buffer, "_cell%d_growthRate_z%d.traj", cell, Zslice);
			thisAddendum = string(buffer, n);
			thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
			thisOutputStream << "\nt = " << thisT << "\n";
			for(int j = 0; j < yl; ++j){
				for(int i = 0; i < xl; ++i){
					thisOutputStream << lattice[idx(i, (yl - j - 1), Zslice, yl, zl)] << "\t";
				}
				thisOutputStream << "\n";
			}
			thisOutputStream.close();
			
			
			//volumeFraction second
			n = sprintf(buffer, "/cells/%d/volumeFraction/%.6E", cell, thisT);
			thisAddendum = string(buffer, n);
			
			thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
			thisInputStream.read((char *)lattice, sizeof(float) * N);
			thisInputStream.close();
			
			n = sprintf(buffer, "_cell%d_volumeFraction_z%d.traj", cell, Zslice);
			thisAddendum = string(buffer, n);
			thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
			thisOutputStream << "\nt = " << thisT << "\n";
			for(int j = 0; j < yl; ++j){
				for(int i = 0; i < xl; ++i){
					thisOutputStream << lattice[idx(i, (yl - j - 1), Zslice, yl, zl)] << "\t";
				}
				thisOutputStream << "\n";
			}
			thisOutputStream.close();
			
			//do the fluxes:
			for(int fbaInp = 0; fbaInp < cells->lattices[cell].fba.nOutputs - 1; ++fbaInp){
				n = sprintf(buffer, "/cells/%d/chem%dflux/%.6E", cell, cells->lattices[cell].fba.outputIndexes[fbaInp], thisT);
				thisAddendum = string(buffer, n);
				
				thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
				thisInputStream.read((char *)lattice, sizeof(float) * N);
				thisInputStream.close();
				
				n = sprintf(buffer, "_cell%d_chem%dflux_z%d.traj", cell, cells->lattices[cell].fba.outputIndexes[fbaInp], Zslice);
				thisAddendum = string(buffer, n);
				thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
				thisOutputStream << "\nt = " << thisT << "\n";
				for(int j = 0; j < yl; ++j){
					for(int i = 0; i < xl; ++i){
						thisOutputStream << lattice[idx(i, (yl - j - 1), Zslice, yl, zl)] << "\t";
					}
					thisOutputStream << "\n";
				}
				thisOutputStream.close();
			}
		}
		
		//chems
		for(int chem = 0; chem < chems->numberChems; ++chem){
			
			n = sprintf(buffer, "/chems/%d/%.6E", chem, thisT);
			thisAddendum = string(buffer, n);
			
			thisInputStream.open((fn + thisAddendum).c_str(), ios_base::binary);
			thisInputStream.read((char *)lattice, sizeof(float) * N);
			thisInputStream.close();
			
			n = sprintf(buffer, "_chem%d_z%d.traj", chem, Zslice);
			thisAddendum = string(buffer, n);
			thisOutputStream.open((fn + thisAddendum).c_str(), ios_base::app);
			thisOutputStream << "\nt = " << thisT << "\n";
			for(int j = 0; j < yl; ++j){
				for(int i = 0; i < xl; ++i){
					thisOutputStream << lattice[idx(i, (yl - j -1), Zslice, yl, zl)] << "\t";
				}
				thisOutputStream << "\n";
			}
			thisOutputStream.close();
		}
		
		thisT += growT * pullbackSteps;
	}
}


//===================================================DIFFUSION KERNELS==============================================================
/// @brief Compute the diffusion of a chemical species
/// @param cells struct containing al the cell lattices
/// @param chems struct containing all the chem lattices
/// @param chemIdx the index of the chemical species that is diffusing
/// @param thisTau the diffusion timestep for this chemical species
/// @return Void
__global__ void chemDiffusion(chemLattices *chems, cellLattices *cells, int chemIdx, float thisTau){
	int yl = chems->yl;
	int zl = chems->zl;
	float lambda = chems->lambda;
	
	int totIndex = (blockIdx.x * gridDim.y + blockIdx.y) * blockDim.x + threadIdx.x;
	
	int i = totIndex / (yl * zl);
	int j = (totIndex % (yl * zl)) / zl;
	int k = (totIndex % (yl * zl)) % zl;
	
	//i would pull a slice of each lattice into shared but I wont yet:
	//__shared__
	
	int thisIdx = idx(i, j, k, yl, zl);
	
	int upIdx = idx(i, j, (k + 1), yl, zl);
	int downIdx = idx(i, j, (k - 1), yl, zl);
	int leftIdx = idx((i - 1), j, k, yl, zl);
	int rightIdx = idx((i + 1), j, k, yl, zl);
	int forwardIdx = idx(i, (j + 1), k, yl, zl);
	int backwardIdx = idx(i, (j - 1), k, yl, zl);
	
	float temp = 0.0f;
	
	int *cTypeFlags = cells->cTypeFlags;
	float *cLattice = chems->lattices[chemIdx].cLattice;
	float *cDlattice = chems->lattices[chemIdx].cDlattice;
	
	float thisConc = cLattice[thisIdx];
	float thisD = cDlattice[thisIdx];
	
	if(cTypeFlags[thisIdx] > 0){
		//sites types
		//0: air - gives but takes nothing
		//1: agar - takes from air, gives nothing to air
		//2: cells - takes from air, takes from agar, gives to agar, gives nothing to air.
		//siteType 1, 2, take only from 0, give and take from 1, 2
		
		temp += ((int)((cTypeFlags[upIdx] > 0) || (cLattice[upIdx] > thisConc))) * (thisD + cDlattice[upIdx]) * (cLattice[upIdx] - thisConc);
		temp += ((int)((cTypeFlags[downIdx] > 0) || (cLattice[downIdx] > thisConc))) * (thisD + cDlattice[downIdx]) * (cLattice[downIdx] - thisConc);
		temp += ((int)((cTypeFlags[leftIdx] > 0) || (cLattice[leftIdx] > thisConc))) * (thisD + cDlattice[leftIdx]) * (cLattice[leftIdx] - thisConc);
		temp += ((int)((cTypeFlags[rightIdx] > 0) || (cLattice[rightIdx] > thisConc))) * (thisD + cDlattice[rightIdx]) * (cLattice[rightIdx] - thisConc);
		temp += ((int)((cTypeFlags[forwardIdx] > 0) || (cLattice[forwardIdx] > thisConc))) * (thisD + cDlattice[forwardIdx]) * (cLattice[forwardIdx] - thisConc);
		temp += ((int)((cTypeFlags[backwardIdx] > 0) || (cLattice[backwardIdx] > thisConc))) * (thisD + cDlattice[backwardIdx]) * (cLattice[backwardIdx] - thisConc);
		
		temp *= ((int)(cTypeFlags[thisIdx] > 0)) * thisTau / (2.0f * lambda * lambda);
		chems->lattices[chemIdx].cTempLattice[thisIdx] = thisConc + temp;
// 		if(!(temp == 0)){
// 		printf("chemType %d, i = %d, j = %d, k = %d, temp is %f, and conc is %f\n", chemIdx, i, j, k, temp, chems->lattices[chemIdx].cTempLattice[thisIdx]);
// 		}
	}
}


//=======================================================REACTION KERNELS===============================================
/// @brief Compute Michaelis-Menten type uptake reactions
/// @param cells struct containing al the cell lattices
/// @param chems struct containing all the chem lattices
/// @param rxns struct containing all the uptake reactions
/// @param thisTau the diffusion timestep for this chemical species
/// @return Void
__global__ void chemReact(chemLattices *chems, cellLattices *cells, chemReactions *rxns){
	int yl = chems->yl;
	int zl = chems->zl;
	
	int totIndex = (blockIdx.x * gridDim.y + blockIdx.y) * blockDim.x + threadIdx.x;
	
	int i = totIndex / (yl * zl);
	int j = (totIndex % (yl * zl)) / zl;
	int k = (totIndex % (yl * zl)) % zl;
	
	//i would pull a slice of each lattice into shared but I wont yet:
	//__shared__
	
	int thisIdx = idx(i, j, k, yl, zl);
	
	for(int i = 0; i < chems->numberChems; ++i){
		chems->lattices[i].cTempLattice[thisIdx] = chems->lattices[i].cLattice[thisIdx]; //sync temp chem lattice with chem lattice locally
	}
	
	if (cells->cTypeFlags[thisIdx] != 2){
		return; //dont do this kernel if there are no cells here
	}
	
	float temp;
	//loop through the reactions:
	for(int i = 0; i < rxns->numberReactions; ++i){
		temp = 0.0;
		temp = cells->lattices[rxns->reactions[i].cellType].cVolumeFraction[thisIdx] //cell volume fraction 
			* rxns->reactions[i].E //times the density of enzymes in a vol fraction
			* rxns->reactions[i].kcat //times kcat
			* chems->lattices[rxns->reactions[i].reactant].cLattice[thisIdx] //times reactant
			* chems->fbaTau
			/ (chems->lattices[rxns->reactions[i].reactant].cLattice[thisIdx] + rxns->reactions[i].km);
			
		chems->lattices[rxns->reactions[i].reactant].cTempLattice[thisIdx] -= temp;
		chems->lattices[rxns->reactions[i].product].cTempLattice[thisIdx] += temp;
	}
}


//======================================================FBA KERNELS======================================================
//TODO: Find out why sometimes a whole bunch of oxygen spontaneously appears---maybe index overflow? try uints?
/// @brief Compute approximate FBA solutions for the cells
/// @param cells struct containing al the cell lattices
/// @param chems struct containing all the chem lattices
/// @return Void
__global__ void roundDownFBA(cellLattices *cells, chemLattices *chems){
	int yl = chems->yl;
	int zl = chems->zl;
	
	int totIndex = (blockIdx.x * gridDim.y + blockIdx.y) * blockDim.x + threadIdx.x;
	
	int i = totIndex / (yl * zl);
	int j = (totIndex % (yl * zl)) / zl;
	int k = (totIndex % (yl * zl)) % zl;
	
	//i would pull a slice of each lattice into shared but I wont yet:
	//__shared__
	
	int thisIdx = idx(i, j, k, yl, zl);
	
	for(int i = 0; i < chems->numberChems; ++i){
		chems->lattices[i].cTempLattice[thisIdx] = chems->lattices[i].cLattice[thisIdx]; //sync temp chem lattice with chem lattice locally
	}
	if(cells->cTypeFlags[thisIdx] == 2){
		float conf = cells->conversionFactorConcPerSecondToMmolPerGDwtPerHr; //has 1/tau already in it
		
		uint thisIndex;
		uint thisPartialIndex;
		uint thisOtherIndex;
		float thisOut;
		float thisOtherOut;
		float outputs[11];
		float deltas[10];
		bool tooBigs[10];
		float locVolFrac;
		
		
		cellLattice thisCellLattice;
		fbaModel thisFbaModel;
		int jthIdx;
		
		for(int i = 0; i < cells->numberCells; ++i){
			thisCellLattice = cells->lattices[i];
			thisFbaModel = thisCellLattice.fba;
			//we are looking at cell type i; we want to find out the appropriate fba solution:
			thisIndex = 0;
			
			//check local vol frac and find if this cell is here
			locVolFrac = thisCellLattice.cVolumeFraction[thisIdx];
			if(locVolFrac == 0.0f){
				continue; // if its not here skip following code
			}
					
			for(int j = 0; j < thisFbaModel.nInputs; ++j){
				//calculate the partial index and the overhang
				jthIdx = thisFbaModel.inputIndexes[j];
				deltas[j] = chems->lattices[jthIdx].cLattice[thisIdx] // [c] / tau
							* conf * (chems->lattices[jthIdx].isIntracellular ? (1.0f / locVolFrac) : 1.0f); // if its passively imported mult available conc by vof frac -- also tau is in conf
									
				thisPartialIndex = (uint)(deltas[j] / thisFbaModel.inputStride[j]);
				deltas[j] -= thisPartialIndex * thisFbaModel.inputStride[j];
				
				//make sure the index is not too big:
				tooBigs[j] = (thisPartialIndex >= thisFbaModel.numberOfTabledValues[j]);
				thisPartialIndex = tooBigs[j] ? (thisFbaModel.numberOfTabledValues[j] - 1) : thisPartialIndex; //if its too big, give back next to last
				
				thisIndex += thisPartialIndex * thisFbaModel.indexMultipliers[j]; //times an index multiplier
			}
			
			for(int j = 0; j < thisFbaModel.nOutputs; ++j){
				//add up each output contribution:
				thisOut = thisFbaModel.cFbaTable[thisIndex * thisFbaModel.nOutputs + j];
				outputs[j] = thisOut;
				
				//calc the change in this output with respect to changes in each of the inputs:
				for(int k = 0; k < thisFbaModel.nInputs; ++k){
					//calc this other index for a change in the kth input
					thisOtherIndex = thisIndex + thisFbaModel.indexMultipliers[k] * (1 - tooBigs[k]); //find the next one in the kth direction
					thisOtherOut = thisFbaModel.cFbaTable[thisOtherIndex * thisFbaModel.nOutputs + j];
					outputs[j] += deltas[k] * (thisOtherOut - thisOut) / thisFbaModel.inputStride[k];
				}
			}
			
			//store the Fba solutions...
			for(int j = 0; j < thisFbaModel.nOutputs; ++j){
				thisCellLattice.cFbaOuts[j][thisIdx] = outputs[j];
			}
			//we have the outputs for cell type i, so make the changes to the chem lattices:
			for(int j = 0; j < thisFbaModel.nOutputs - 1; ++j){
				chems->lattices[thisFbaModel.outputIndexes[j]].cTempLattice[thisIdx] += (outputs[j] / conf) * locVolFrac; //this FORWARDS convention has NEGATIVE EXCHANGE going INTO the cell!!!!!!!, and make sure this site is cell type
				
			}
		}
	}
}

/// @brief Grow the cells
/// @param cells struct containing al the cell lattices
/// @param growT the growth time
/// @return Void
__global__ void growCells(cellLattices *cells, float growT){
	//grow the cells, no need to swap vol fracs afterward, and cGr is deleted
	int yl = cells->yl;
	int zl = cells->zl;
	
	int totIndex = (blockIdx.x * gridDim.y + blockIdx.y) * blockDim.x + threadIdx.x;
	
	int i = totIndex / (yl * zl);
	int j = (totIndex % (yl * zl)) / zl;
	int k = (totIndex % (yl * zl)) % zl;
	
	//i would pull a slice of each lattice into shared but I wont yet:
	//__shared__
	cellLattice thisCellLattice;
	int thisIdx = idx(i, j, k, yl, zl);
	for(int i = 0; i < cells->numberCells; ++i){
		thisCellLattice = cells->lattices[i];
		thisCellLattice.cVolumeFraction[thisIdx] = thisCellLattice.cVolumeFraction[thisIdx] * __expf(growT * thisCellLattice.cFbaOuts[thisCellLattice.fba.nOutputs - 1][thisIdx] / 3600.0f);
	}
}


//======================================================CELL EXPANSION KERNELS===============================================

/// @brief Expand the cells
/// @param cells struct containing al the cell lattices
/// @param chems struct containing all the chem lattices
/// @return Void
__global__ void expandCells(cellLattices *cells, chemLattices *chems){
	//SYNC CELLS AND CHEMS BEFORE RUNNING THIS KERNEL!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	int yl = cells->yl;
	int zl = cells->zl;
	
	int totIndex = (blockIdx.x * gridDim.y + blockIdx.y) * blockDim.x + threadIdx.x;
	
	int i = totIndex / (yl * zl);
	int j = (totIndex % (yl * zl)) / zl;
	int k = (totIndex % (yl * zl)) % zl;
	
	//i would pull a slice of each lattice into shared but I wont yet:
	//__shared__
	
	int thisIdx = idx(i, j, k, yl, zl);
	
	//in this i need to subtract the amount that may leave -- IF it leaves -- and take in the amount that comes in -- IF it comes in.
	
	int indicies[6];
	indicies[0] = idx(i, j, (k + 1), yl, zl);
	indicies[1] = idx(i, j, (k - 1), yl, zl);
	indicies[2] = idx((i - 1), j, k, yl, zl);
	indicies[3] = idx((i + 1), j, k, yl, zl);
	indicies[4] = idx(i, (j + 1), k, yl, zl);
	indicies[5] = idx(i, (j - 1), k, yl, zl);
	
	float max = cells->maxVolFraction;
	
	float thisTVF = cells->cTotalVolumeFraction[thisIdx];
	int thisTvfZero = (int)(thisTVF <= 0.0f);
	float otherTVF;
	int otherTvfZero;
	
	float thisP = thisTVF - max;
	thisP *= ((int)(thisP > 0.0f));
	
	//can this be a cell?
	int thisCanBeCell = (int)(cells->cTypeFlags[thisIdx] != 1);
	int nnCanBeCell;
		
	float otherP;
	float amnt;
	float temp;
	int inOut;
	
	cellLattice thisCellLattice;
	chemLattice thisChemLattice;
	
	//up first, then down, left, right, forward, backward:
	for(int nn = 0; nn < 6; ++nn){
		//can the neighbor be a cell?
		nnCanBeCell = (int)(cells->cTypeFlags[indicies[nn]] != 1);
		
		//figure out how much of the total amnt should move through each face
		otherTVF = cells->cTotalVolumeFraction[indicies[nn]];
		otherP =  otherTVF - max;
		otherP *= ((int)(otherP > 0.0f));		
		amnt = (thisP - otherP) / 12.0f; //12 = 2 * 6 = half of the sixth that should pass each face into the nearest neighbor
		inOut = ((int)(amnt < 0.0f)); //if inOut = 1, the flow is into this subvolume:
		
		otherTvfZero = (int)(otherTVF <= 0.0f);
		
		//flow in the cell vols
		for(int i = 0; i < cells->numberCells; ++i){
			thisCellLattice = cells->lattices[i];
			temp = 0.0f;
			temp += (thisCellLattice.cVolumeFraction[indicies[nn]] / (otherTVF + otherTvfZero)) * inOut;
			temp += (thisCellLattice.cVolumeFraction[thisIdx] / (thisTVF + thisTvfZero)) * (1 - inOut);
			temp *= amnt * thisCanBeCell * nnCanBeCell;
			thisCellLattice.cTempVolumeFraction[thisIdx] -= temp;
			
			for(int j = 0; j < chems->numberChems; ++j){
				thisChemLattice = chems->lattices[j];
				temp = 0.0f;
				temp += (thisCellLattice.cVolumeFraction[indicies[nn]] / (otherTVF + otherTvfZero)) * inOut * thisChemLattice.cLattice[indicies[nn]];
				temp += (thisCellLattice.cVolumeFraction[thisIdx] / (thisTVF + thisTvfZero)) * (1 - inOut) * thisChemLattice.cLattice[thisIdx];
				temp *= amnt * thisCanBeCell * nnCanBeCell;
				thisChemLattice.cTempLattice[thisIdx] -= temp * thisChemLattice.isIntracellular * (int)(thisChemLattice.cellIdxChemLivesIn == i);
			}
		}
	}
}


//=====================================================UPDATE D KERNELS=================================================
/// @brief compute diffusion ratesand flags everywhere
/// @param cells struct containing al the cell lattices
/// @param chems struct containing all the chem lattices
/// @return Void
__global__ void calcDandFlags(cellLattices *cells, chemLattices *chems){
	//write out D matrix for the chemical
	
	int yl = cells->yl;
	int zl = cells->zl;
	
	int totIndex = (blockIdx.x * gridDim.y + blockIdx.y) * blockDim.x + threadIdx.x;
	
	int i = totIndex / (yl * zl);
	int j = (totIndex % (yl * zl)) / zl;
	int k = (totIndex % (yl * zl)) % zl;
	
	int thisIdx = idx(i, j, k, yl, zl);
	
	int airOrAgar = ((int)(k > chems->agarHeight));
	int cellsOrNot = ((int)(cells->cTotalVolumeFraction[thisIdx] > 0.0f));
	int isHindered;
	
	float temp = (1.0f - cells->cTotalVolumeFraction[thisIdx]) / (1.0f + 0.5f * cells->cTotalVolumeFraction[thisIdx]);
	temp *= ((int)(temp > 0.0f));
	
	chemLattice thisChemLattice;
	
	for(int l = 0; l < chems->numberChems; ++l){
		thisChemLattice = chems->lattices[l];
		isHindered = thisChemLattice.isHindered;
		thisChemLattice.cDlattice[thisIdx] =
												(1 - airOrAgar) * thisChemLattice.Dagar //
												+ airOrAgar * (1 - cellsOrNot) * thisChemLattice.Dwater
												+ airOrAgar * cellsOrNot * (1 - isHindered) * thisChemLattice.Dwater
												+ airOrAgar * cellsOrNot * isHindered * temp * thisChemLattice.Dwater;
	}
	
	cells->cTypeFlags[thisIdx] = 
								(1 - airOrAgar) * 1
								+ airOrAgar * (1 - cellsOrNot) * 0
								+ airOrAgar * cellsOrNot * 2;
}


//====================================================Cell Regulaton Kernels===========================================
/// @brief Run regulation reactions on the cells
/// @param cells struct containing al the cell lattices
/// @param chems struct containing all the chem lattices
/// @param regs struct containing all the regulation reaction structs
/// @param growT the growth time
/// @return Void
__global__ void runRegulation(cellLattices *cells, chemLattices *chems, regReactions *regs, float growT){
	//SYNC CELLS BEFORE YOU RUN THIS KERNELL!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	//SYNC CHEMS BEFORE YOU RUN THIS KERNEL!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
	//the reason you need to sync is that on iteration of the loop can edit the current amount of cellTypeXX if its a reactant, but then
	//if its a product for a later type that edit will be undone :(
	
	//FIGURE OUT A WAY TO KEEP INTERNAL STUFF INSIDE THE CELLS AS THEY CHANGE TYPE!!!!
	//FOR EXAMPLE INTERNAL GLUCOSE IN TYPE0 SHOULD TURN TO INTERNAL GLUCOSE IN TYPE1 AS CELL0 GOES TO CELL1
	
	int yl = cells->yl;
	int zl = cells->zl;
	
	int totIndex = (blockIdx.x * gridDim.y + blockIdx.y) * blockDim.x + threadIdx.x;
	
	int i = totIndex / (yl * zl);
	int j = (totIndex % (yl * zl)) / zl;
	int k = (totIndex % (yl * zl)) % zl;
	
	int thisIdx = idx(i, j, k, yl, zl);
	
	float thisRate;
	float thisChemRate;
	
	float x;
	float y;
	float xx;
	float yy;
	float xy;
	float xxx;
	float yyy;
	float xxy;
	float xyy;
	
	for(int reg = 0; reg < regs->numberReactions; ++reg){
		//we look at each reg reaction;
		//we find the mults:
		x = chems->lattices[regs->reactions[reg].chemType1].cLattice[thisIdx];
		y = chems->lattices[regs->reactions[reg].chemType2].cLattice[thisIdx];
		xx = x * x;
		yy = y * y;
		xy = x * y;
		xxx = xx * x;
		yyy = yy * y;
		xxy = xx * y;
		xyy = x * yy;
		
		//we find the rate:
		thisRate = regs->reactions[reg].multipliers[0]
					+ regs->reactions[reg].multipliers[1] * x
					+ regs->reactions[reg].multipliers[2] * y
					+ regs->reactions[reg].multipliers[3] * xx
					+ regs->reactions[reg].multipliers[4] * yy
					+ regs->reactions[reg].multipliers[5] * xy
					+ regs->reactions[reg].multipliers[6] * xxx
					+ regs->reactions[reg].multipliers[7] * yyy
					+ regs->reactions[reg].multipliers[8] * xxy
					+ regs->reactions[reg].multipliers[9] * xyy;
		thisRate *= (int)(thisRate > 0.0f) * cells->lattices[regs->reactions[reg].reactantCell].cVolumeFraction[thisIdx] * growT;
		
		cells->lattices[regs->reactions[reg].reactantCell].cTempVolumeFraction[thisIdx] -= thisRate;
		cells->lattices[regs->reactions[reg].productCell].cTempVolumeFraction[thisIdx] += thisRate;
		
		for(int cog = 0; cog < regs->reactions[reg].numChemCognates; ++cog){
			thisChemRate = thisRate * chems->lattices[regs->reactions[reg].cognates[cog].reactChemIdx].cLattice[thisIdx];
			chems->lattices[regs->reactions[reg].cognates[cog].reactChemIdx].cTempLattice[thisIdx] -= thisChemRate;
			chems->lattices[regs->reactions[reg].cognates[cog].prodChemIdx].cTempLattice[thisIdx] += thisChemRate;
		}
	}
}


//=====================================================SUPPORT KERNELS=======================================================

/// @brief impose reflective oundary conditions on the simulation walls for a given chem
/// @param chems struct containing all the chem lattices
/// @param chem index of the chosen chem to have reflective boundary conditions
/// @return Void
__global__ void reflectiveWallBC(chemLattices *chems, int chem){
	
	int xl = chems->xl;
	int yl = chems->yl;
	int zl = chems->zl;
	int locInd = blockIdx.x * blockDim.x + threadIdx.x;
	int xPj;
	int xPk;
	int yPi;
	int yPk;
	
	chemLattice thisChemLattice = chems->lattices[chem];
	//Do the x = {0, L} planes first
	if(locInd < (yl * zl)){
		xPj = locInd / zl;
		xPk = locInd % zl;
		
		thisChemLattice.cLattice[idx(-1, xPj, xPk, yl, zl)] = thisChemLattice.cLattice[idx(0, xPj, xPk, yl, zl)];
		thisChemLattice.cLattice[idx(xl, xPj, xPk, yl, zl)] = thisChemLattice.cLattice[idx((xl - 1), xPj, xPk, yl, zl)];
	}
	//Do the y = {0, L} planes next
	if(locInd < (xl * zl)){
		yPi = locInd / zl;
		yPk = locInd % zl;
		
		thisChemLattice.cLattice[idx(yPi, -1, yPk, yl, zl)] = thisChemLattice.cLattice[idx(yPi, 0, yPk, yl, zl)];
		thisChemLattice.cLattice[idx(yPi, yl, yPk, yl, zl)] = thisChemLattice.cLattice[idx(yPi, (yl - 1), yPk, yl, zl)];
	}
}

/// @brief impose reflective boundary conditions on the simulation floor for a given chem
/// @param chems struct containing all the chem lattices
/// @param chem index of the chosen chem to have reflective boundary conditions
/// @return Void
__global__ void reflectiveFloorBC(chemLattices *chems, int chem){
	int xl = chems->xl;
	int yl = chems->yl;
	int zl = chems->zl;
	int locInd = blockIdx.x * blockDim.x + threadIdx.x;
	int zPi;
	int zPj;
	
	chemLattice thisChemLattice = chems->lattices[chem];
	//Do the z = 0 plane
	if(locInd < (xl * yl)){
		zPi = locInd / yl;
		zPj = locInd % yl;
		
		thisChemLattice.cLattice[idx(zPi, zPj, -1, yl, zl)] = thisChemLattice.cLattice[idx(zPi, zPj, 0, yl, zl)];
	}
}

/// @brief resum the total volume fraction of all the cell types
/// @param cells struct containing all the cell lattices
/// @return Void
__global__ void resumTotalVolFract(cellLattices *cells){
	int yl = cells->yl;
	int zl = cells->zl;
	
	int totIndex = (blockIdx.x * gridDim.y + blockIdx.y) * blockDim.x + threadIdx.x;
	
	int i = totIndex / (yl * zl);
	int j = (totIndex % (yl * zl)) / zl;
	int k = (totIndex % (yl * zl)) % zl;
	
	//i would pull a slice of each lattice into shared but I wont yet:
	//__shared__
	
	int thisIdx = idx(i, j, k, yl, zl);
	
	cells->cTotalVolumeFraction[thisIdx] = 0.0f;
	for(int i = 0; i < cells->numberCells; ++i){
		cells->cTotalVolumeFraction[thisIdx] += cells->lattices[i].cVolumeFraction[thisIdx];
	}
}


/// @brief copy volume fraction to temp volume fraction lattice
/// @param cells struct containing all the cell lattices
/// @return Void
__global__ void syncTempVolFractWithVolFract(cellLattices *cells){
	int yl = cells->yl;
	int zl = cells->zl;
	
	int totIndex = (blockIdx.x * gridDim.y + blockIdx.y) * blockDim.x + threadIdx.x;
	
	int i = totIndex / (yl * zl);
	int j = (totIndex % (yl * zl)) / zl;
	int k = (totIndex % (yl * zl)) % zl;
	
	//i would pull a slice of each lattice into shared but I wont yet:
	//__shared__
	
	int thisIdx = idx(i, j, k, yl, zl);
	
	for(int i = 0; i < cells->numberCells; ++i){
		cells->lattices[i].cTempVolumeFraction[thisIdx] = cells->lattices[i].cVolumeFraction[thisIdx];
	}
}


/// @brief copy chem to temp chem lattice
/// @param chems struct containing all the chem lattices
/// @return Void
__global__ void syncTempChemLatticeWithChemLattice(chemLattices *chems){
	int yl = chems->yl;
	int zl = chems->zl;
	
	int totIndex = (blockIdx.x * gridDim.y + blockIdx.y) * blockDim.x + threadIdx.x;
	
	int i = totIndex / (yl * zl);
	int j = (totIndex % (yl * zl)) / zl;
	int k = (totIndex % (yl * zl)) % zl;
	
	//i would pull a slice of each lattice into shared but I wont yet:
	//__shared__
	
	int thisIdx = idx(i, j, k, yl, zl);
	
	for(int i = 0; i < chems->numberChems; ++i){
		chems->lattices[i].cTempLattice[thisIdx] = chems->lattices[i].cLattice[thisIdx];
	}
}


/// @brief swap cell volume fraction lattices and temp volume fraction lattices
/// @param cells struct containing all the cell lattices
/// @return Void
__global__ void swapCellLattices(cellLattices *cells){
	//float *temp;
	for(int i = 0; i < cells->numberCells; ++i){
		float *temp = cells->lattices[i].cTempVolumeFraction;
		cells->lattices[i].cTempVolumeFraction = cells->lattices[i].cVolumeFraction;
		cells->lattices[i].cVolumeFraction = temp;
	}
}


/// @brief swap a single cell volume fraction lattice and the corresponding temp volume fraction lattice
/// @param cells struct containing all the cell lattices
/// @param cellIdx cell index
/// @return Void
__global__ void swapACellLattice(cellLattices *cells, int cellIdx){
	float *temp;
	temp = cells->lattices[cellIdx].cTempVolumeFraction;
	cells->lattices[cellIdx].cTempVolumeFraction = cells->lattices[cellIdx].cVolumeFraction;
	cells->lattices[cellIdx].cVolumeFraction = temp;
}


/// @brief swap chem lattices and temp chem lattices
/// @param chems struct containing all the chem lattices
/// @return Void
__global__ void swapChemLattices(chemLattices *chems){
	//float *temp;
	for(int i = 0; i < chems->numberChems; ++i){
		float *temp = chems->lattices[i].cTempLattice;
		chems->lattices[i].cTempLattice = chems->lattices[i].cLattice;
		chems->lattices[i].cLattice = temp;
	}
}


/// @brief swap a single cell volume fraction lattice and the corresponding temp volume fraction lattice
/// @param cells struct containing all the cell lattices
/// @param cellIdx cell index
/// @return Void
__global__ void swapAChemLattice(chemLattices *chems, int chemIdx){
	float *temp;
	temp = chems->lattices[chemIdx].cTempLattice;
	chems->lattices[chemIdx].cTempLattice = chems->lattices[chemIdx].cLattice;
	chems->lattices[chemIdx].cLattice = temp;

}


#define relaxIdx2d(j, k, zl) (j * zl + k)
/// @brief first of three reduction kernels for checking if cells are relaxed
/// @param cells struct containing all the cell lattices
/// @param tempX0plane 2D slice the size of YxZ
/// @return Void
__global__ void checkIsRelaxed1(cellLattices *cells, float *tempX0plane){
	//drillDown:
	int xl = cells->xl;
	int yl = cells->yl;
	int zl = cells->zl;
	
	int j = (blockIdx.x * blockDim.x + threadIdx.x) / zl;
	int k = (blockIdx.x * blockDim.x + threadIdx.x) % zl;
	
	tempX0plane[relaxIdx2d(j, k, zl)] = 0.0f;
	
	for(int i = 0; i < xl; ++i){
		tempX0plane[relaxIdx2d(j, k, zl)] = cells->cTotalVolumeFraction[idx(i, j, k, yl, zl)] > tempX0plane[relaxIdx2d(j, k, zl)] ?
											cells->cTotalVolumeFraction[idx(i, j, k, yl, zl)] : tempX0plane[relaxIdx2d(j, k, zl)];
	}
}

/// @brief second of three reduction kernels for checking if cells are relaxed
/// @param cells struct containing all the cell lattices
/// @param tempX0plane 2D slice the size of YxZ
/// @param tempX0Y0line 1D slice (line) the size of Z
/// @return Void
__global__ void checkIsRelaxed2(cellLattices *cells, float *tempX0plane, float *tempX0Y0line){
	//drillAcross:
	int yl = cells->yl;
	int zl = cells->zl;
	
	int k = threadIdx.x;
	
	tempX0Y0line[k] = 0.0f;
	
	for(int j = 0; j < yl; ++j){
		tempX0Y0line[k] = tempX0plane[relaxIdx2d(j, k, zl)] > tempX0Y0line[k] ?
						tempX0plane[relaxIdx2d(j, k, zl)] : tempX0Y0line[k];
	}
}

/// @brief first of three reduction kernels for checking if cells are relaxed
/// @param cells struct containing all the cell lattices
/// @param tempX0Y0line 1D slice (line) the size of Z
/// @return Void
__global__ void checkIsRelaxed3(float *tempX0Y0line, cellLattices *cells){
	// drill through
	int zl = cells->zl;
	
	float max = 0.0f;
	
	for(int k = 0; k < zl; ++k){
		max = tempX0Y0line[k] > max ? tempX0Y0line[k] : max;
	}
	
	cells->isRelaxed = (max - cells->maxVolFraction) > maxVolFracTolerance ? 0 : 1;
}

/// @brief function that checks if the cells are relaxed, it calls the checkIsRelaxed1, 2, and 3
/// @param cells struct containing all the cell lattices
/// @param cCells device struct containing all the cell lattices
/// @param tempX0plane 2D slice the size of YxZ
/// @param tempX0Y0line 1D slice (line) the size of Z
/// @return Void
void checkIsRelaxed(cellLattices *cells, cellLattices *cCells, float *tempX0plane, float *tempX0Y0line){
	int yl = cells->yl;
	int zl = cells->zl;
	
	int numBlocks = yl * zl / defaultThreads;
	checkIsRelaxed1<<<numBlocks, defaultThreads>>>(cCells, tempX0plane);
	checkIsRelaxed2<<<1, zl>>>(cCells, tempX0plane, tempX0Y0line);
	checkIsRelaxed3<<<1, 1>>>(tempX0Y0line, cCells);
	cudaMemcpy(&(cells->isRelaxed), &(cCells->isRelaxed), sizeof(int), cudaMemcpyDeviceToHost);

}

/// @brief free up all the cuda memory
/// @param cells struct containing all the cell lattices
/// @param chems struct containing all the chem lattices
/// @param cCells device struct containing all the cell lattices
/// @param cChems device struct containing all the chem lattices
/// @param cRxns device struct containing all the rxns
/// @param cRegs device struct containing all the regulation reactions
/// @return Void
void freeCudaMemory(cellLattices *cells, chemLattices *chems, cellLattices *cCells, chemLattices *cChems, chemReactions *cRxns, regReactions *cRegs){
	//free the device stuff
	cudaFree(cCells);
	cudaFree(cChems);
	cudaFree(cRxns);
	cudaFree(cRegs);
	
	//free the cells stuff
	for(int cell = 0; cell < cells->numberCells; ++cell){
		cudaFree(cells->lattices[cell].cVolumeFraction);
		cudaFree(cells->lattices[cell].cTempVolumeFraction);
		//cudaFree(cells->lattices[cell].cFbaIndexes);
		for(int i = 0; i < cells->lattices[cell].fba.nOutputs; ++i){
			cudaFree(cells->lattices[cell].cFbaOuts[i]);
		}
		cudaFree(cells->lattices[cell].fba.cFbaTable);
	}
	cudaFree(cells->cTypeFlags);
	cudaFree(cells->cTotalVolumeFraction);
	
	//free the chems stuff
	for(int chem = 0; chem < chems->numberChems; ++chem){
		cudaFree(chems->lattices[chem].cLattice);
		cudaFree(chems->lattices[chem].cTempLattice);
		cudaFree(chems->lattices[chem].cDlattice);
	}
	
}

//====================================================Container Functions for chem relaxation and growth===============
/// @brief container function that relaxes the chem profiles to equilibrium.  This calls the diffusion, uptake reaction, and fba kernels
/// @param cells struct containing all the cell lattices
/// @param cCells device struct containing all the cell lattices
/// @param chems struct containing all the chem lattices
/// @param cChems device struct containing all the chem lattices
/// @param cRxns device struct containing all the rxns
/// @param timesteps number of timesteps to run for
/// @param thisBlockDim block dimensions
/// @return Void
void runChemRelaxation(cellLattices *cells, cellLattices *cCells, chemLattices *chems, chemLattices *cChems, chemReactions *cRxns, int timesteps, blockDims thisBlockDim){
	
	// TODO:  remove dependency on chems->lattices[chem].tau
	int numDiffusions;
	float dTau;
	
	
	for(int i = 0; i < timesteps; ++i){
		//run diffusion steps:
		for(int chem = 0; chem < chems->numberChems; ++chem){
			if(chems->lattices[chem].isIntracellular == 0){
				numDiffusions = (int)(chems->fbaTau / chems->lattices[chem].tau) + 1;
				dTau = chems->fbaTau / numDiffusions;
				for(int j = 0; j < numDiffusions; ++j){
					//boundary conditions
					if(chems->lattices[chem].refWallBcFlag == 1){
						reflectiveWallBC<<<thisBlockDim.BCblocksWall, defaultThreads>>>(cChems, chem);
					}
					if(chems->lattices[chem].refFloorBcFlag == 1){
						reflectiveFloorBC<<<thisBlockDim.BCblocksFloor, defaultThreads>>>(cChems, chem);
					}
					//diffusion kernel
					chemDiffusion<<<thisBlockDim.diffNumBlocks, diffThreads>>>(cChems, cCells, chem, dTau);
					//swap
					swapAChemLattice<<<1, 1>>>(cChems, chem);
					swapAChemLattice_host(chems, chem);
				}
			}
		}
		
 		//run reaction steps:
 		//syncTempChemLatticeWithChemLattice<<<thisBlockDim.syncTCLwithCLNumBlocks, syncTCLwithCLThreads>>>(cChems);
 		chemReact<<<thisBlockDim.chemReactNumBlocks, chemReactThreads>>>(cChems, cCells, cRxns);
 		swapChemLattices<<<1, 1>>>(cChems);
 		swapChemLattices_host(chems);
		
		//run FBA on chems:
		//syncTempChemLatticeWithChemLattice<<<thisBlockDim.syncTCLwithCLNumBlocks, syncTCLwithCLThreads>>>(cChems);
		roundDownFBA<<<thisBlockDim.fbaNumBlocks, fbaThreads>>>(cCells, cChems);
		swapChemLattices<<<1, 1>>>(cChems);
		swapChemLattices_host(chems);
		
	}
}

/// @brief container function that relaxes the chem profiles to equilibrium.  This calls the diffusion, uptake reaction, and fba kernels
/// @param cells struct containing all the cell lattices
/// @param cCells device struct containing all the cell lattices
/// @param chems struct containing all the chem lattices
/// @param cChems device struct containing all the chem lattices
/// @param cRegs device struct containing all the regulation reactions
/// @param growT the growth time
/// @param tempX0plane 2D slice the size of YxZ
/// @param tempX0Y0line 1D slice (line) the size of Z
/// @param thisBlockDim block dimensions
/// @return Void
void runCellGrowth(chemLattices *chems, chemLattices *cChems, cellLattices *cells, cellLattices *cCells, regReactions *cRegs, float growT, float *tempX0plane, float *tempX0Y0line, blockDims thisBlockDim){

	//using the stored grs, grow the cells:
 	growCells<<<thisBlockDim.defaultNumBlocks, defaultThreads>>>(cCells, growT);
 				
 	//resum the total vol fraction:
 	resumTotalVolFract<<<thisBlockDim.defaultNumBlocks, defaultThreads>>>(cCells);
	
	while(true){
 		syncTempChemLatticeWithChemLattice<<<thisBlockDim.syncTCLwithCLNumBlocks, syncTCLwithCLThreads>>>(cChems);
 		syncTempVolFractWithVolFract<<<thisBlockDim.defaultNumBlocks, defaultThreads>>>(cCells);
		expandCells<<<thisBlockDim.expandCellsNumBlocks, expandCellsThreads>>>(cCells, cChems);
 		swapCellLattices<<<1, 1>>>(cCells);
		swapCellLattices_host(cells);
 		swapChemLattices<<<1, 1>>>(cChems);
		swapChemLattices_host(chems);
		
		//resum total volume fraction
		resumTotalVolFract<<<thisBlockDim.defaultNumBlocks, defaultThreads>>>(cCells);
		
		checkIsRelaxed(cells, cCells, tempX0plane, tempX0Y0line);
		if(cells->isRelaxed == 1){
			break;
		}	
 	}
			
	// run gene regulation -- this was moved based on a fear that the regulation would be adding SMALL 
	// numbers iteratively that may get swamped out due to rounding errors.  Instead the regulation
	// is now only performed when the cells grow -- this is not treated on equal footing as growth
	// because growth rates are solved each fba iteration and averaged in the end; this is just estimating 
	// the correct regulation based on the most current chemical state, there is no averaging:
 	syncTempVolFractWithVolFract<<<thisBlockDim.defaultNumBlocks, defaultThreads>>>(cCells);
 	syncTempChemLatticeWithChemLattice<<<thisBlockDim.syncTCLwithCLNumBlocks, syncTCLwithCLThreads>>>(cChems);
 	runRegulation<<<thisBlockDim.defaultNumBlocks, defaultThreads>>>(cCells, cChems, cRegs, growT);
 	swapCellLattices<<<1, 1>>>(cCells);
 	swapCellLattices_host(cells);
 	swapChemLattices<<<1, 1>>>(cChems);
 	swapChemLattices_host(chems);
 			
 	//update D and Flags:
 	calcDandFlags<<<thisBlockDim.defaultNumBlocks, defaultThreads>>>(cCells, cChems);
}

//====================================================MAIN=============================================================

int main(int argc, char* argv[]){
	
	if(argc == 1){
		cout << "Proper Usage:\ncontinuumColony -i <Input Filename>" << endl << endl;
		cout << "optional flags:" << endl;
		cout << "-d <device ID> specifies device" << endl;
		cout << "-r specifies restart" << endl;
		cout << "-Xslice <int> writes trajectory files for x = int slice of lattice" << endl;
		cout << "-Yslice <int> writes trajectory files for y = int slice of lattice" << endl;
		cout << "-Zslice <int> writes trajectory files for z = int slice of lattice" << endl;
		cout << endl << "NOTE: to make output trajectories once a simulation is finished use both -r and -<XYZ>slice" << endl;
		exit(-1);
	}
	
	
	//read command line data
	string filename;
	int restart = 0;
	int Xslice = -1;
	int Yslice = -1;
	int Zslice = -1;
	
	for (int argCounter = 1; argCounter < argc; ++argCounter){
		if(string(argv[argCounter]).compare("-i") == 0){
			filename = string(argv[argCounter + 1]);
			cout << "accessing " << filename << "..." << endl;
		}
		if(string(argv[argCounter]).compare("-d") == 0){
			int dev = (int)atoi(argv[argCounter + 1]);
			cudaSetDevice(dev);
			cout << "setting device number " << dev << "..." << endl;
		}
		if(string(argv[argCounter]).compare("-r") == 0){
			restart = 1;
			cout << "this is a restart..." << endl;
		}
		if(string(argv[argCounter]).compare("-Xslice") == 0){
			Xslice = (int)atoi(argv[argCounter + 1]);
			cout << "Xslice = " << Xslice << " specified..." << endl;
		}
		if(string(argv[argCounter]).compare("-Yslice") == 0){
			Yslice = (int)atoi(argv[argCounter + 1]);
			cout << "Yslice = " << Yslice << " specified..." << endl;
		}
		if(string(argv[argCounter]).compare("-Zslice") == 0){
			Zslice = (int)atoi(argv[argCounter + 1]);
			cout << "Zslice = " << Zslice << " specified..." << endl;
		}
	}
	
	
	// read simulation size, and prepare blocks and such...
	int xl = getLx(filename);
	int yl = getLy(filename);
	int zl = getLz(filename);
	
	blockDims thisBlockDim;
	thisBlockDim.makeBlockDims(xl, yl, zl);
	
	
	if(!(xl * yl * zl % defaultThreads == 0)){
		cerr << "ERROR: go back and define your volume such that X x Y x Z is a multiple of 512..." << endl;
		exit(-3);
	}
	if(!(yl * zl % defaultThreads == 0)){
		cerr << "ERROR: go back and define your volume such that Y x Z is a multiple of 512..." << endl;
		exit(-3);
	}
	if(!(zl < defaultThreads)){
		cerr << "ERROR: go back and define your volume such that Z is less than 512..." << endl;
		exit(-3);
	}
	if(!(xl * yl * zl % chemReactThreads == 0)){
		cerr << "ERROR: go back and define your volume such that X x Y x Z is a multiple of " << chemReactThreads << "..." << endl;
		exit(-3);
	}
	
	//read other useful info from input file
	float lambda = getLambda(filename);
	float fbaTau = getFbaTau(filename);
	float cellVolume = getCellVolume(filename);
	float cellMass = getCellMass(filename);
	int agarHeight = getAgarHeight(filename);
	int maxCellNumber = 0;
	
	//creat my container structs:
	cellLattices cells;
	cells.buildCellLattices(getMaxVolFrac(filename), xl, yl, zl, lambda, fbaTau, agarHeight, cellVolume, cellMass);
	cout << "built cell lattices..." << endl;
	
	chemLattices chems;
	chems.buildChemLattices(xl, yl, zl, lambda, fbaTau, agarHeight);
	cout << "built chem lattices..." << endl;
	
	chemReactions rxns;
	cout << "built rxns..." << endl;
	
	regReactions regs;
	cout << "built regs..." << endl;
	
	//allocate cuda versions:
	cellLattices *cCells;
	cudaMalloc((void **)&cCells, sizeof(cellLattices));
	
	chemLattices *cChems;
	cudaMalloc((void **)&cChems, sizeof(chemLattices));
	
	chemReactions *cRxns;
	cudaMalloc((void **)&cRxns, sizeof(regReactions));
	
	regReactions *cRegs;
	cudaMalloc((void **)&cRegs, sizeof(regReactions));
	
	cout << "allocated device structs..." << endl;
	
	//set up the structs
	while(true){
		//add the cell types
		if(isCellHere(filename, cells)){
			fbaModel workingModel;
			workingModel.buildFbaModel(getFbaFilename(filename, cells));
			cellLattice workingCell;
			int startNumber = getStartNumber(filename, cells);
			//get the max number of cells in starting config so as to relax enough.
			if(maxCellNumber < startNumber){
				 maxCellNumber = startNumber;
			}
			workingCell.buildCellLattice(workingModel, xl, yl, zl, lambda, getStarterFlag(filename, cells), getArbitraryVolumeFractionFilename(filename, cells), getStartX(filename, cells), getStartY(filename, cells), startNumber, agarHeight, cellVolume);
			cells.add(workingCell);
		}
		else{
			break;
		}
	}
	cout << "set up cell lattices..." << endl;
	
	while(true){
		//add the chem types
		if(isChemHere(filename, chems)){
			chemLattice workingChem;
			workingChem.buildChemLattice(getDwater(filename, chems), getDagar(filename, chems), lambda, getTau(filename, chems), xl, yl, zl, agarHeight, getInitialAgarConc(filename, chems), getInitialAirConc(filename, chems), getIntracellularFlag(filename, chems), getCellIdxChemLivesInFlag(filename, chems), getIsHinderedFlag(filename, chems), cellVolume, getRefWallBcFlag(filename, chems), getRefFloorBcFlag(filename, chems));
			
			chems.add(workingChem);
		}
		else{
			break;
		}
	}
	cout << "set up chem lattices..." << endl;
	
	while(true){
		//add the reactions
		if(isRxnHere(filename, rxns)){
			chemReaction workingRxn;
			workingRxn.buildChemReaction(getKm(filename, rxns), getKcat(filename, rxns), getEnzymeNumberPerCell(filename, rxns), getCellTypeThatDoesThisReaction(filename, rxns), getReactantChemIndex(filename, rxns), getProductChemIndex(filename, rxns), cellVolume);
			
			rxns.add(workingRxn);
		}
		else{
			break;
		}
	}
	cout << "set up rxns..." << endl;
	
	while(true){
		//add the regulations
		if(isRegHere(filename, regs)){
			regReaction workingReg;
			workingReg.buildRegReaction(getRegChemOne(filename, regs), getRegChemTwo(filename, regs), getRegReactCell(filename, regs), getRegProdCell(filename, regs), getRegMultipliersString(filename, regs));
			while(true){
				//add cognates
				if(isCogHere(filename, regs, workingReg)){
					workingReg.fillChemCognate(getChemCognate(filename, regs, workingReg));
				}
				else{
					break;
				}
			}
					
			regs.add(workingReg);
		
		}
		else{
			break;
		}
	}
	cout << "set up regs..." << endl;
	
	//Copy structs over to device:
	cudaMemcpy(cCells, &cells, sizeof(cellLattices), cudaMemcpyHostToDevice);
	cudaMemcpy(cChems, &chems, sizeof(chemLattices), cudaMemcpyHostToDevice);
	cudaMemcpy(cRxns, &rxns, sizeof(chemReactions), cudaMemcpyHostToDevice);
	cudaMemcpy(cRegs, &regs, sizeof(regReactions), cudaMemcpyHostToDevice);
	cout << "copied data to device structs..." << endl;
	
	// do the simulations
	int t = 0;
	float maxT = getMaxT(filename);
	float pullbackT = getPullbackT(filename);
	float growT = getGrowT(filename);
	float checkpointT = getCheckpointT(filename);
	if(checkpointT < 0){
		checkpointT = maxT;
	}
	float chemEqT = getChemEqT(filename);
	float initialChemEqT = getInitialChemEqT(filename);
	
	int pullbackSteps = (int)(pullbackT / growT + 0.5f);
	int checkpointSteps = (int)(checkpointT / growT + 0.5f);
	
	int chemEqSteps = (int)(chemEqT / fbaTau);
	int initialChemEqSteps = (int)(initialChemEqT / fbaTau);
	
	//These are for checking if the cells are relaxed
	float *tempX0plane;
	float *tempX0Y0line;
	cudaMalloc((void **)&tempX0plane, sizeof(float) * yl * zl);
	cudaMalloc((void **)&tempX0Y0line, sizeof(float) * zl);
	
	
	//either relax initial configuration, or else load restart data
	if(restart == 0){
		
		//set up output directory:
		creatNewOutputDirectory(filename, &cells, &chems);
		
		//relax the initial cell configuration:
		cout << "relaxing initial configuration..." << endl;
		
		//resum the total vol fraction:
		resumTotalVolFract<<<thisBlockDim.defaultNumBlocks, defaultThreads>>>(cCells);

		while(true){
			for(int relax10 = 0; relax10 < 10; ++relax10){
					
				//expand the cells:
				syncTempChemLatticeWithChemLattice<<<thisBlockDim.syncTCLwithCLNumBlocks, syncTCLwithCLThreads>>>(cChems);
				syncTempVolFractWithVolFract<<<thisBlockDim.defaultNumBlocks, defaultThreads>>>(cCells);
				expandCells<<<thisBlockDim.expandCellsNumBlocks, expandCellsThreads>>>(cCells, cChems);
				swapCellLattices<<<1, 1>>>(cCells);
				swapCellLattices_host(&cells);
				swapChemLattices<<<1, 1>>>(cChems);
				swapChemLattices_host(&chems);
				
				//resum total vol fraction:
				resumTotalVolFract<<<thisBlockDim.defaultNumBlocks, defaultThreads>>>(cCells);

			}
			checkIsRelaxed(&cells, cCells, tempX0plane, tempX0Y0line);
			if(cells.isRelaxed == 1){
				break;
			}
		}
		
		//update D and Flags:
		calcDandFlags<<<thisBlockDim.defaultNumBlocks, defaultThreads>>>(cCells, cChems);
		
		//bring chems to equilibrium:
		cout << "bringing chemistry to initial equilibrium..." << endl;
		runChemRelaxation(&cells, cCells, &chems, cChems, cRxns, initialChemEqSteps, thisBlockDim);
		
		//write zero timepoint
		cellMemCopyBack(&cells);
		chemMemCopyBack(&chems);
		
		writeToExisting(filename, &cells, &chems, 0.0f);
	}
	else{
		//load data from restart file:
		cout << "loading data from restart files:" << endl;
		//TODO load data!!!!!!!!!!!!!!!
		openExistingZippedDirectory(filename);
		float chkPtT = restartFromExisting(filename, &cells, &chems);
		
		t = (int)(chkPtT / growT + 0.5f);
		
		cellMemCopyToDev(&cells);
		chemMemCopyToDev(&chems);
		
		//update D and Flags:
		calcDandFlags<<<thisBlockDim.defaultNumBlocks, defaultThreads>>>(cCells, cChems);
		
	}
	
	//get the simulation going:
	cout << "starting the simulation..." << endl;
	
	int maxSteps = (int)(maxT / growT + 0.5f);
	while(t < maxSteps){
		t++;
		
		//grow the colony for growT:
		runCellGrowth(&chems, cChems, &cells, cCells, cRegs, growT, tempX0plane, tempX0Y0line, thisBlockDim);
		
		//relax the chem concentrations:
		runChemRelaxation(&cells, cCells, &chems, cChems, cRxns, chemEqSteps, thisBlockDim);
		
		if((t % pullbackSteps) == 0){
			cellMemCopyBack(&cells);
			chemMemCopyBack(&chems);
			
			writeToExisting(filename, &cells, &chems, t * growT);
			cout << "current simulation time: " << t * growT << endl;
		}
		
		if((t % checkpointSteps) == 0){
			closeAndZipDirectory(filename);
			cout << "archiving output data..." << endl;
		}
	}
	
	//make trajectories
	if(Xslice != -1){
		cout << "writing Xslice = " << Xslice << " trajectory..." << endl;
		writeYZTrajectory(filename, &cells, &chems, Xslice, growT, pullbackSteps);
	}
	if(Yslice != -1){
		cout << "writing Yslice = " << Yslice << " trajectory..." << endl;
		writeXZTrajectory(filename, &cells, &chems, Yslice, growT, pullbackSteps);
	}
	if(Zslice != -1){
		cout << "writing Zslice = " << Zslice << " trajectory..." << endl;
		writeXYTrajectory(filename, &cells, &chems, Zslice, growT, pullbackSteps);
	}
	
	//free the cuda memory:
	freeCudaMemory(&cells, &chems, cCells, cChems, cRxns, cRegs);
	
	closeAndZipDirectory(filename);
	string finalCommand = "rm -rf " + filename.substr(0, filename.find("."));
	system(finalCommand.c_str());
	
	return 0;
}
